import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class JungeWeltParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath(
            "//div[@class = 'row']/div[contains(@class, 'col') and not(@class = 'col-md-8 mx-auto mt-4 bg-light')]/p"
        )
        _summary_selector = CSSSelector(".teaser.lead")
        _subheadline_selector = XPath("//div[@class = 'row']/div[contains(@class,'col')]/h3")
        _free_access_selector = XPath("//h1[text()='Sie sind nun eingeloggt.']|//p[@class='m-1']")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("Author"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute
        def free_access(self) -> bool:
            return not bool(self._free_access_selector(self.precomputed.doc))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                caption_selector=XPath("./ancestor::figure//div[contains(@class, 'caption')]"),
                relative_urls=True,
            )



from fundus import Requires
from fundus.scraping.filter import RequiresAll


class TestExtractionFilter:
    def test_requires(self):
        extraction = {"a": "Some Stuff", "b": [], "c": True}

        assert not Requires("a")(extraction)

        assert (result := Requires("a", "b")(extraction))

        assert result.missing_attributes == ("b",)

        assert not Requires("c")(extraction)

        extraction = {"a": "Some Stuff", "b": [], "c": False}

        assert (result := Requires("a", "b", "c")(extraction))

        assert sorted(result.missing_attributes) == sorted(("b", "c"))

        assert not Requires("c", eval_booleans=False)(extraction)

    def test_requires_all(self):
        extraction = {"a": "Some Stuff", "b": [], "c": False}

        assert (result := RequiresAll()(extraction))
        assert result.missing_attributes == ("b",)

        extraction = {"a": "Some Stuff", "c": False}
        assert not RequiresAll()(extraction)

        # test skip_boolean=False
        extraction = {"a": "Some Stuff", "b": [], "c": False}

        assert (result := RequiresAll(eval_booleans=True)(extraction))
        assert sorted(result.missing_attributes) == sorted(("b", "c"))

        extraction = {"a": "Some Stuff", "c": True}
        assert not RequiresAll(eval_booleans=True)(extraction)



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    apply_substitution_pattern_over_list,
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class TokyoChunichiShimbunParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath("//main//div[@class='block' and not(descendant::div or descendant::h2)]")
        _subheadline_selector = XPath("//main//div[@class='block']//h2")

        _author_bloat_pattern = re.compile(r"記者")
        _topic_bloat_pattern = re.compile(r"話題・|話題")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def authors(self) -> List[str]:
            return apply_substitution_pattern_over_list(
                generic_author_parsing(self.precomputed.ld.bf_search("author")), self._author_bloat_pattern
            )

        @attribute
        def topics(self) -> List[str]:
            if topics := apply_substitution_pattern_over_list(
                generic_topic_parsing(self.precomputed.ld.bf_search("articleSection")), self._topic_bloat_pattern
            ):
                return [topic for topic in topics if "ニュース" not in topic]
            return []

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=CSSSelector("main div.image img, main div.thumb img"),
                caption_selector=XPath(
                    "./ancestor::div[@class='wrap']//p[@class='caption'] | "
                    "./ancestor::div[@class='thumb']//p[@class='thumb-caption']"
                ),
                author_selector=re.compile(r"（(?P<credits>[^）]*?)(撮影)?）\s*$"),
                relative_urls=True,
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.url import RSSFeed, Sitemap

from .morgunbladid import MorgunbladidParser

# noinspection PyPep8Naming


class ISL(metaclass=PublisherGroup):
    default_language = "is"

    Morgunbladid = Publisher(
        name="Morgunbladid",
        domain="https://www.mbl.is/",
        parser=MorgunbladidParser,
        sources=[
            Sitemap("https://www.mbl.is/mm/sitemap.xml"),
            RSSFeed("https://www.mbl.is/feeds/fp/"),
            RSSFeed("https://www.mbl.is/feeds/innlent/"),
            RSSFeed("https://www.mbl.is/feeds/erlent/"),
            RSSFeed("https://www.mbl.is/feeds/english/"),
            RSSFeed("https://www.mbl.is/feeds/helst/"),
        ],
    )



from __future__ import annotations

import html
import itertools
import json
import re
from collections import defaultdict
from copy import copy
from dataclasses import dataclass, field
from datetime import datetime
from functools import total_ordering
from typing import (
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Iterator,
    List,
    Match,
    NamedTuple,
    Optional,
    Pattern,
    Sequence,
    Set,
    Type,
    Union,
    cast,
)
from urllib.parse import urljoin

import lxml.html
import more_itertools
import validators
from dateutil import parser
from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.logging import create_logger
from fundus.parser.data import (
    DOM,
    ArticleBody,
    ArticleSection,
    Dimension,
    Image,
    ImageVersion,
    LinkedDataMapping,
    TextSequence,
)
from fundus.utils.regex import _get_match_dict
from fundus.utils.serialization import JSONVal

logger = create_logger(__name__)

_space_characters = {
    "whitespace": r"\s",
    "non-breaking-space": r"\u00A0",
    "zero-width-space": r"\u200B",
    "zero-width-non-joiner": r"\u200C",
    "zero-width-joiner": r"\u200D",
    "zero-width-no-break_space": r"\uFEFF",
}
_ws_pattern: Pattern[str] = re.compile(rf'[{"".join(_space_characters.values())}]+')


def normalize_whitespace(text: str) -> str:
    return re.sub(_ws_pattern, " ", text).strip()


@total_ordering
@dataclass(eq=False)
class Node:
    position: int
    node: lxml.html.HtmlElement = field(compare=False)
    _break_selector: ClassVar[XPath] = XPath("*//br")

    # one could replace this recursion with XPath using an expression like this:
    # //*[not(self::script) and text()]/text(), but for whatever reason, that's actually 50-150% slower
    # than simply using the implemented mixture below
    def text_content(self, excluded_tags: Optional[List[str]] = None, tag_filter: Optional[XPath] = None) -> str:
        guarded_excluded_tags: List[str] = excluded_tags or []

        def _text_content(element: lxml.html.HtmlElement) -> str:
            if (
                element.tag in guarded_excluded_tags
                or isinstance(element, lxml.html.HtmlComment)
                or (tag_filter and tag_filter(element))
            ):
                return element.tail or ""
            text = element.text or ""
            children = "".join([_text_content(child) for child in element.iterchildren()])
            tail = element.tail or ""
            return text + children + tail

        return _text_content(self._get_break_preserved_node())

    def _get_break_preserved_node(self) -> lxml.html.HtmlElement:
        copied_node = copy(self.node)
        for br in self._break_selector(copied_node):
            br.tail = "\n" + br.tail if br.tail else "\n"
        return copied_node

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Node):
            return NotImplemented
        return self.position == other.position

    def __lt__(self, other: "Node") -> bool:
        return self.position < other.position

    def __hash__(self) -> int:
        return hash(self.position)

    def __str__(self) -> str:
        return self.text_content()

    def __repr__(self):
        return f"{type(self).__name__}: {self.text_content()}"

    def __bool__(self):
        return bool(normalize_whitespace(self.text_content()))


class SummaryNode(Node):
    pass


class SubheadNode(Node):
    pass


class ParagraphNode(Node):
    pass


def extract_article_body_with_selector(
    doc: lxml.html.HtmlElement,
    paragraph_selector: XPath,
    summary_selector: Optional[XPath] = None,
    subheadline_selector: Optional[XPath] = None,
    tag_filter: Optional[XPath] = None,
) -> ArticleBody:
    # depth first index for each element in tree
    df_idx_by_ref = {element: i for i, element in enumerate(doc.iter())}

    def extract_nodes(selector: XPath, node_type: Type[Node]) -> List[Node]:
        if not selector and node_type:
            raise ValueError("Both a selector and node type are required")

        return [node for element in selector(doc) if (node := node_type(df_idx_by_ref[element], element))]

    summary_nodes = extract_nodes(summary_selector, SummaryNode) if summary_selector else []
    subhead_nodes = extract_nodes(subheadline_selector, SubheadNode) if subheadline_selector else []
    paragraph_nodes = extract_nodes(paragraph_selector, ParagraphNode)
    nodes = sorted(summary_nodes + subhead_nodes + paragraph_nodes)

    if not nodes:
        # return empty body if no text is present
        return ArticleBody(TextSequence([]), [])

    instructions = more_itertools.split_when(nodes, pred=lambda x, y: type(x) != type(y))

    if not summary_nodes:
        instructions = more_itertools.prepend([], instructions)
    elif not nodes[: len(summary_nodes)] == summary_nodes:
        raise ValueError(f"All summary nodes should be at the beginning of the article")

    if not subhead_nodes or (paragraph_nodes and subhead_nodes[0] > paragraph_nodes[0]):
        first = next(instructions)
        instructions = itertools.chain([first, []], instructions)

    summary = TextSequence(
        map(
            lambda x: normalize_whitespace(x.text_content(excluded_tags=["script"], tag_filter=tag_filter)),
            next(instructions),
        )
    )
    sections: List[ArticleSection] = []

    for chunk in more_itertools.chunked(instructions, 2):
        if len(chunk) == 1:
            chunk.append([])
        texts = [
            list(
                map(lambda x: normalize_whitespace(x.text_content(excluded_tags=["script"], tag_filter=tag_filter)), c)
            )
            for c in chunk
        ]
        sections.append(ArticleSection(*map(TextSequence, texts)))

    return ArticleBody(summary=summary, sections=sections)


_ld_node_selector = XPath("//script[@type='application/ld+json']")
_json_pattern = re.compile(r"(?P<json>{[\s\S]*}|\[\s*{[\s\S]*}\s*](?!\s*}))")
_json_undefined = re.compile(r'(?P<key>"[^"]*?"):\s*undefined')


def sanitize_json(text: str) -> Optional[str]:
    # capture only content enclosed as follows: {...} or [{...}]
    match = re.search(_json_pattern, text)
    if match is None or not (sanitized := match.group("json")):
        return None

    # substitute "bad" values
    sanitized = re.sub(_json_undefined, r"\g<key>:null", sanitized)
    sanitized = re.sub(r"[\r\n\t]+", "", sanitized)
    return sanitized


def unescape_json_values(obj):
    if isinstance(obj, str):
        return html.unescape(obj)
    elif isinstance(obj, list):
        return [unescape_json_values(item) for item in obj]
    elif isinstance(obj, dict):
        return {key: unescape_json_values(value) for key, value in obj.items()}
    else:
        return obj


def parse_json(text: str) -> Optional[Dict[str, JSONVal]]:
    if not (json_content := sanitize_json(text)):
        return None

    try:
        parsed_json = json.loads(json_content)
        return cast(Dict[str, JSONVal], unescape_json_values(parsed_json))
    except json.JSONDecodeError as error:
        logger.debug(f"Encountered {error!r} during JSON parsing")
        return None


def extract_json_from_dom(root: lxml.html.HtmlElement, selector: XPath) -> Iterable[Dict[str, JSONVal]]:
    jsons = [parsed_json for node in selector(root) if (parsed_json := parse_json(node.text_content())) is not None]
    return more_itertools.collapse(jsons, base_type=dict)


def get_ld_content(root: lxml.html.HtmlElement) -> LinkedDataMapping:
    """Parse JSON-LD from HTML.

    This function parses a script tags of type ld+json.
    In case the JSON is wrapped in a CDATA tag it is first stripped.

    Args:
        root: The HTML document given as a lxml.html.HtmlElement.

    Returns:
        The JSON-LD data as a LinkedDataMapping
    """

    return LinkedDataMapping(extract_json_from_dom(root, _ld_node_selector))


_meta_node_selector = CSSSelector("head > meta, body > meta, article > meta")


def get_meta_content(root: lxml.html.HtmlElement) -> Dict[str, str]:
    """Parse metadata from HTML.

    This function parses single values (i.e. charset=...), nodes containing name, property, http-equiv or
    itemprop attributes. When multiple values for the same key occur, they will be joined using `,`. This
    is in order to ease typing and avoid list as additional type.

    In case an HTML tag consists a class, it will be appended as namespace to avoid key collisions.
    I.e. <meta class="swiftype" name="author" ... > will be stored using `swiftype:author` as a key.

    Args:
        root: The HTML document given as a lxml.html.HtmlElement.

    Returns:
        The metadata as a dictionary
    """
    data = defaultdict(list)
    for node in _meta_node_selector(root):
        attributes = node.attrib
        if len(attributes) == 1:
            data[attributes.keys()[0]].append(attributes.values()[0])
        elif key := (  # these keys are ordered by frequency
            attributes.get("name")
            or attributes.get("property")
            or attributes.get("http-equiv")
            or attributes.get("itemprop")
        ):
            if ns := attributes.get("class"):
                key = f"{ns}:{key}"
            if content := attributes.get("content"):
                data[key].append(content)

    metadata: Dict[str, str] = {}
    for name, listed_content in data.items():
        if len(listed_content) == 1:
            metadata[name] = listed_content[0]
        else:
            # for ease of typing we join multiple contents for the same key using ','
            metadata[name] = ",".join(listed_content)

    return metadata


def transform_breaks_to_paragraphs(element: lxml.html.HtmlElement, **attribs: str) -> lxml.html.HtmlElement:
    """Splits the content of <element> on <br> tags into paragraphs and transform them in <p> elements.

    Args:
        element: The element on which to perform the transformation
        **attribs: The attributes of the wrapped paragraphs as keyword arguments. I.e. the
            default {"class": "br-wrap"} wil produce the following elements: <p class='br-wrap'>.
            To use python keywords wrap them dunder scores. __class__ for class.

    Returns:
        The transformed element
    """

    if not attribs:
        attribs = {"class": "br-wrap"}
    else:
        attribs = {re.sub(r"^__(.*?)__$", r"\1", key): value for key, value in attribs.items()}

    def get_paragraphs() -> List[str]:
        raw_html = lxml.etree.tostring(element, method="html", encoding="unicode")
        if match := re.match(r"^<[^>]*?>\s*(?P<content>.*?)\s*<[^>]*?>\s*$", raw_html, re.S):
            content = match.group("content")
            return list(filter(bool, (text.strip() for text in content.split("<br>"))))
        return []

    def generate_attrs() -> str:
        return " ".join([f"{attribute}='{value}'" for attribute, value in attribs.items()]) if attribs else ""

    def clear_element():
        for child in element:
            element.remove(child)
        element.tail = None
        element.text = None

    # split content on <br> tags
    if not (paragraphs := get_paragraphs()):
        return element

    # remove children, tail and text from element
    clear_element()

    # add paragraphs to cleared element
    for paragraph in paragraphs:
        wrapped = f"<p{' ' + generate_attrs()}>{paragraph}</p>"
        element.append(lxml.html.fromstring(wrapped))

    return element


def strip_nodes_to_text(text_nodes: List[lxml.html.HtmlElement], join_on: str = "\n\n") -> Optional[str]:
    if not text_nodes:
        return None
    return join_on.join(([re.sub(r"\n+", " ", node.text_content()) for node in text_nodes])).strip()


def generic_nodes_to_text(nodes: Sequence[Union[lxml.html.HtmlElement, str]], normalize: bool = False) -> List[str]:
    if not nodes:
        return []

    texts = []
    for node in nodes:
        if isinstance(node, lxml.html.HtmlElement):
            text = str(node.text_content())
        elif isinstance(node, str):
            text = node
        else:
            raise TypeError(f"Unexpected type {type(node)}")

        if normalize:
            text = normalize_whitespace(text)

        if text:
            texts.append(text)

    return texts


def apply_substitution_pattern_over_list(
    input_list: List[str], pattern: Pattern[str], replacement: Union[str, Callable[[Match[str]], str]] = ""
) -> List[str]:
    return [subbed for text in input_list if (subbed := re.sub(pattern, replacement, text).strip())]


def generic_author_parsing(
    value: Union[
        Optional[str],
        Dict[str, str],
        List[str],
        List[Dict[str, str]],
    ],
    split_on: Optional[List[str]] = None,
    normalize: bool = True,
) -> List[str]:
    """This function tries to parse the given <value> to a list of authors (List[str]) based on the type of value.

    Parses based on type of <value> as following:
        value (None):       Empty list \n
        value (str):        re.split(delimiters) with delimiters := split_on or common delimiters' \n
        value (dict):       If key "name" in dict, [dict["name"]] else empty list \n
        value (list[str]):  value\n
        value (list[dict]): [dict["name"] for dict in list if dict["name"]] \n

    with common delimiters := [",", ";", " und ", " and ", " & ", " | "]

    All values are stripped with default strip() method before returned.

    Args:
        value: An input value representing author(s) which get parsed based on type.
        split_on: Only relevant for type(<value>) = str. If set, split <value> on <split_on>,
            else (default) split <value> on common delimiters.
        normalize: If True, normalize every author with normalize_whitespace(). Defaults to True

    Returns:
        A parsed and striped list of authors
    """

    common_delimiters = [",", ";", " und ", " and ", " & ", r" \| "]

    parameter_type_error: TypeError = TypeError(
        f"<value> '{value}' has an unsupported type {type(value)}. "
        f"Supported types are 'Optional[str], Dict[str, str], List[str], List[Dict[str, str]],'"
    )

    def parse_author_dict(author_dict: Dict[str, str]) -> Optional[str]:
        if (author_name := author_dict.get("name")) is not None:
            return author_name

        given_name = author_dict.get("givenName", "")
        additional_name = author_dict.get("additionalName", "")
        family_name = author_dict.get("familyName", "")
        if given_name and family_name:
            return " ".join(filter(bool, [given_name, additional_name, family_name]))
        else:
            return None

    if not value:
        return []

    # collapse
    authors: List[str] = []

    for item in value if isinstance(value, list) else [value]:
        if isinstance(item, str):
            authors.append(item)

        elif isinstance(item, dict):
            if (author := parse_author_dict(item)) is not None:
                authors.append(author)

        else:
            raise parameter_type_error

    if normalize or split_on:

        def split(text: str) -> Iterable[str]:
            return filter(bool, re.split(r"|".join(split_on or common_delimiters), text))

        authors = list(more_itertools.flatten([split(author) for author in authors]))
        normalized_authors = [normalize_whitespace(author) for author in authors]
        return normalized_authors

    return authors


def generic_text_extraction_with_css(doc, selector: XPath) -> Optional[str]:
    nodes = selector(doc)
    return strip_nodes_to_text(nodes)


def generic_topic_parsing(
    keywords: Optional[Union[str, List[str]]], delimiter: Union[str, List[str]] = ","
) -> List[str]:
    if isinstance(delimiter, str):
        delimiter = [delimiter]

    if not keywords:
        topics = []
    elif isinstance(keywords, str):
        topics = [
            cleaned
            for keyword in re.split(pattern=f"[{''.join(delimiter)}]", string=keywords)
            if (cleaned := keyword.strip())
        ]
    elif isinstance(keywords, list) and all(isinstance(s, str) for s in keywords):
        topics = keywords
    else:
        raise TypeError(f"Encountered unexpected type {type(keywords)} as keyword parameter")

    return list(dict.fromkeys(topics))


_tz_infos = {"CET": 3600, "CEST": 7200, "IST": 19800}


class CustomParserInfo(parser.parserinfo):
    MONTHS = [
        ("Jan", "January", "Januar"),
        ("Feb", "February", "Februar"),
        ("Mar", "March", "März"),
        ("Apr", "April"),
        ("May", "May", "Mai"),
        ("Jun", "June", "Juni"),
        ("Jul", "July", "Juli"),
        ("Aug", "August"),
        ("Sep", "Sept", "September"),
        ("Oct", "October", "Oktober", "Okt"),
        ("Nov", "November"),
        ("Dec", "December", "Dezember", "Dez"),
    ]


def generic_date_parsing(date_str: Optional[str]) -> Optional[datetime]:
    return parser.parse(date_str, tzinfos=_tz_infos, parserinfo=CustomParserInfo(), fuzzy=True) if date_str else None


_title_selector = CSSSelector("title")


def parse_title_from_root(root: lxml.html.HtmlElement) -> Optional[str]:
    title_node = _title_selector(root)

    if len(title_node) != 1:
        return None

    return strip_nodes_to_text(title_node)


def preprocess_url(url: str, domain: str) -> str:
    url = re.sub(r"\\/", "/", url)
    # Some publishers use relative URLs
    if not validators.url(url):
        publisher_domain = "https://" + domain
        url = urljoin(publisher_domain, url)
    return url


def image_author_parsing(authors: Union[str, List[str]]) -> List[str]:
    credit_keywords = [
        "fotograf",
        "credits?",
        "quellen?",
        "bild(rechte)?",
        "sources?",
        r"(((f|ph)oto(graph)?s?|image|illustrations?|cartoons?|pictures?)\s*)+(by|:|courtesy)",
        "©",
        "– alle rechte vorbehalten",
        "copyright",
        "all rights reserved",
        "courtesy of",
        "＝",
    ]
    author_filter = re.compile(r"(?is)^(" + r"|".join(credit_keywords) + r"):?\s*")

    def clean(author: str):
        author = re.sub(r"^\((.*)\)$", r"\1", author).strip()
        # filtering credit keywords
        author = re.sub(author_filter, "", author, count=1)
        # filtering bloat follwing the author
        author = re.sub(r"(?i)/?copyright.*", "", author)
        return author.strip()

    if isinstance(authors, list):
        authors = [clean(author) for author in authors]
    else:
        authors = clean(authors)
    return generic_author_parsing(authors)


# https://regex101.com/r/MplUXL/2
_srcset_pattern = re.compile(r"(?P<url>\S+)\s*(?P<descriptor>[0-9.]+[wx])?(,?\s*)")


def parse_srcset(srcset: str) -> Dict[str, str]:
    # Updated regular expression to account for query parameters in URLs
    urls = {}
    for match in _srcset_pattern.finditer(srcset.strip()):
        url = match.group("url")
        descriptor = match.group("descriptor")  # Width (w) or pixel density (x)
        urls[descriptor or "1x"] = url
    # return sorted dict based on int value of descriptor
    return dict(sorted(urls.items(), key=lambda item: float(item[0][:-1])))


# that's the same as string(./attribute::*[ends-with(name(), '*')]) but LXML doesn't support the ends-with function
# these two selectors select the value of the first attribute found ending with src/srcset relative to the node
# as truing value
_srcset_selector = XPath(
    "./@*[substring(name(), string-length(name()) - string-length('srcset') + 1)  = 'srcset'][starts-with(., 'http') or starts-with(., '/')]"
)
_src_selector = XPath(
    "./@*[substring(name(), string-length(name()) - string-length('src') + 1)  = 'src'][starts-with(., 'http') or starts-with(., '/')]"
)


def parse_urls(node: lxml.html.HtmlElement) -> Optional[Dict[str, str]]:
    def get_longest_string(strings: List[str]) -> str:
        return sorted(strings, key=len)[-1]

    if srcset := cast(List[str], _srcset_selector(node)):
        return parse_srcset(get_longest_string(srcset))
    elif src := cast(List[str], _src_selector(node)):
        return {"1x": get_longest_string(src)}
    else:
        return None


class _DimensionCalculator:
    def __init__(
        self, width: Optional[float] = None, height: Optional[float] = None, ratio: Optional[float] = None
    ) -> None:
        self.width = width
        self.height = height
        self.ratio = ratio

    def calculate(
        self, width: Optional[float] = None, height: Optional[float] = None, dpr: Optional[float] = None
    ) -> Optional[Dimension]:
        if not (width or height):
            width = self.width
            height = self.height
        if dimension := Dimension.from_ratio(width, height, self.ratio):
            return dimension * (dpr or 1)
        return None


_media_param_pattern = re.compile(r"\(\s*(?P<param>[\w-]+)\s*:\s*(?P<value>[\d./]+)(?P<unit>[a-z]*)\)")
_width_x_height_pattern = re.compile(r"(?P<width>[0-9]+)x(?P<height>[0-9]+)")


def get_versions_from_node(
    source: lxml.html.HtmlElement, ratio: Optional[float], size_pattern: Optional[Pattern[str]]
) -> Set[ImageVersion]:
    if not (urls := parse_urls(source)):
        return set()

    # get min/max width
    query_width = None
    for param, value, descriptor in re.findall(_media_param_pattern, source.get("media", "").split(",")[0]):
        if param in ["min-width", "max-width"]:
            if descriptor != "px":
                logger.debug(f"Pixel calculation not implemented for {descriptor}")
            else:
                # with the assumption that there is only one max/min width per ',' seperated query and only
                # either min- or max-width
                query_width = f"{param}:{value}"

    # get width, height and init calculator
    if (src_width := source.get("width")) and src_width.replace(".", "", 1).isdigit():
        width = float(src_width or 0) or None
    else:
        width = None
    if (src_height := source.get("height")) and src_height.replace(".", "", 1).isdigit():
        height = float(src_height or 0) or None
    else:
        height = None
    if width and height:
        ratio = width / height
    calculator = _DimensionCalculator(width, height, ratio)

    versions = set()
    for descriptor, url in urls.items():
        kwargs: Dict[str, float] = {}
        if descriptor is not None:
            if match := re.search(r"(?P<multiplier>[0-9.]+)x", descriptor):
                kwargs["dpr"] = float(match.group("multiplier"))
            elif match := re.search(r"(?P<width>[0-9]+)(px|w)", descriptor):
                kwargs["width"] = float(match.group("width"))

        if size_pattern is not None and (
            match_dict := _get_match_dict(size_pattern, url, conversion=lambda x: float(x))
        ):
            kwargs.update(match_dict)
        elif not (calculator.width or kwargs.get("width")) and (match := re.search(_width_x_height_pattern, url)):
            kwargs.update({k: float(v) for k, v in match.groupdict().items() if v is not None})

        version = ImageVersion(
            url=url, query_width=query_width, size=calculator.calculate(**kwargs), type=source.get("type")
        )
        versions.add(version)

    return versions


_relative_source_selector = XPath("./ancestor::picture//source")


def parse_versions(img_node: lxml.html.HtmlElement, size_pattern: Optional[Pattern[str]] = None) -> List[ImageVersion]:
    # parse img
    if (
        (default_width := img_node.get("width"))
        and not default_width == "auto"
        and (default_height := img_node.get("height"))
        and not default_height == "auto"
    ):
        ratio = float(default_width) / float(default_height)
    else:
        ratio = None

    versions = set()
    for source in _relative_source_selector(img_node) + [img_node]:
        for version in get_versions_from_node(source, ratio, size_pattern):
            versions.add(version)

    return sorted(versions)


class IndexedImageNode(NamedTuple):
    position: int
    content: lxml.html.HtmlElement
    is_cover: bool


def parse_image_nodes(
    image_nodes: List[IndexedImageNode],
    caption_selector: XPath,
    alt_selector: XPath,
    author_selector: Union[XPath, Pattern[str]],
    domain: Optional[str] = None,
    size_pattern: Optional[Pattern[str]] = None,
) -> Iterator[Image]:
    """Extract urls, caption, description and authors from a list of <img> nodes

    Args:
        image_nodes: Indexed <img> nodes to parse.
        caption_selector: Selector selecting the caption of an image. Defaults to selecting the figcaption element.
        alt_selector: Selector selecting the descriptive text of an image. Defaults to selecting alt value.
        author_selector: Selector selecting the credits for an image. Defaults to selecting an arbitrary child of
            figure with copyright or credit in its class attribute.
        domain: If set, the domain will be prepended to URLs in case they are relative
        size_pattern: Regular expression to select <width>, <height> and <dpr> from the image URL. The given regExp
            will be matched with re.findall and overwrites existing values. Defaults to None.

    Returns:
        List of Images
    """

    def nodes_to_text(nodes: List[Union[lxml.html.HtmlElement, str]]) -> Optional[str]:
        return " ".join(generic_nodes_to_text(nodes, normalize=True)) or None

    for position, node, is_cover in image_nodes:
        # parse URLs
        if not (versions := parse_versions(node, size_pattern)):
            continue

        # resolve relative URLs if domain is given
        if domain is not None:
            for version in versions:
                version.url = urljoin(domain, version.url)

        # parse caption
        caption = nodes_to_text(caption_selector(node))

        # parse description
        description = nodes_to_text(alt_selector(node))

        # parse authors
        authors = []
        if isinstance(author_selector, Pattern):
            # author is part of the caption
            if caption and (match := re.search(author_selector, caption)):
                authors = [match.group("credits")]
                caption = re.sub(author_selector, "", caption).strip() or None
            elif description and (match := re.search(author_selector, description)):
                authors = [match.group("credits")]
        else:
            # author is selectable as node
            if author_nodes := author_selector(node):
                authors = generic_nodes_to_text(author_nodes, normalize=True)
        authors = image_author_parsing(authors)

        yield Image(
            versions=versions,
            caption=caption,
            authors=authors,
            description=description,
            is_cover=is_cover,
            position=position,
        )


class Bounds(NamedTuple):
    upper: int
    first_paragraph: Optional[int]
    lower: int


def determine_bounds(
    dom: DOM, paragraph_selector: XPath, upper_boundary_selector: XPath, lower_boundary_selector: Optional[XPath]
) -> Optional[Bounds]:
    def get_sorted_indices(nodes: List[lxml.html.HtmlElement]) -> List[int]:
        return sorted([dom.get_index(node) for node in nodes])

    # the getitem on upper_boundary_selector ensures that this throws an exception, if there are no
    # upper_boundary_node present, as well as removing excess ones.
    upper_boundary_nodes = [upper_boundary_selector(dom.root)[0]]
    paragraph_nodes = paragraph_selector(dom.root)
    lower_boundary_nodes = lower_boundary_selector(dom.root) if lower_boundary_selector else []

    sorted_indices = get_sorted_indices(upper_boundary_nodes + paragraph_nodes + lower_boundary_nodes)

    if len(sorted_indices) < 2:
        return None

    return Bounds(
        upper=sorted_indices[0],
        first_paragraph=paragraph_indices[0] if (paragraph_indices := get_sorted_indices(paragraph_nodes)) else None,
        lower=sorted_indices[-1],
    )


_og_url_selector = XPath("string(//meta[@property='og:url']/@content)")


def image_extraction(
    doc: lxml.html.HtmlElement,
    paragraph_selector: XPath,
    image_selector: XPath = XPath("//figure//img"),
    upper_boundary_selector: XPath = XPath("//main"),
    lower_boundary_selector: Optional[XPath] = None,
    caption_selector: XPath = XPath("./ancestor::figure//figcaption"),
    alt_selector: XPath = XPath("./@alt"),
    author_selector: Union[XPath, Pattern[str]] = XPath(
        "(./ancestor::figure//*[(contains(@class, 'copyright') or contains(@class, 'credit')) and text()])[1]"
    ),
    relative_urls: Union[bool, XPath] = False,
    size_pattern: Pattern[str] = re.compile(
        r"width([=-])(?P<width>[0-9.]+)|height([=-])(?P<height>[0-9.]+)|dpr=(?P<dpr>[0-9.]+|)"
    ),
) -> List[Image]:
    """Extracts images enriched with metadata from <dom> based on given selectors.

    The core idea behind this function is to select all images matching <image_selector> that lay between
    the first element selected by <upper_boundary_selector> and the last element of <lower_boundary_selector>
    or <paragraph_selector>. The hierarchy is determined by indexing all nodes of <doc> depth first.
    To enrich the selected images with metadata like 'caption', 'alt-description' or `authors`, one should make
    use of the corresponding selectors.

    Args:
        doc: The html document of the article.
        paragraph_selector: Selector used to select the paragraphs of the article.
        image_selector: Selector selecting all relevant img elements. Defaults '//figure//img'.
        upper_boundary_selector: A selector referencing an element to be considered as the upper boundary. All img
            elements before this element will be ignored.
        lower_boundary_selector: A selector referencing an element to be considered as the lower boundary. All img
            elements after this element will be ignored. Defaults to the last paragraph of an article.
        caption_selector: Selector selecting the caption of an image. Defaults to selecting the figcaption element.
        alt_selector: Selector selecting the descriptive text of an image. Defaults to selecting alt value.
        author_selector: Selector selecting the credits for an image. Defaults to selecting an arbitrary child of
            figure with copyright or credit in its class attribute.
        relative_urls: If True, the extractor assumes that image src URLs are relative and prepends the publisher
            domain
        size_pattern: Regular expression to select <width>, <height> and <dpr> from the image URL. The given regExp
            will be matched with re.findall and overwrites existing values. Defaults to None.

    Returns:
        A list of Images contained within the article

    """

    # index nodes df
    dom = DOM(doc)

    # determine bounds based on df index
    if not (bounds := determine_bounds(dom, paragraph_selector, upper_boundary_selector, lower_boundary_selector)):
        raise ValueError("Bounds could not be determined")

    if relative_urls:
        if isinstance(relative_urls, bool):
            selector = _og_url_selector
        else:
            selector = relative_urls
        if not (domain := selector(dom.root)):
            raise ValueError("Could not determine domain")
    else:
        domain = None

    image_nodes = [
        IndexedImageNode(position=position, content=node, is_cover=position < (bounds.first_paragraph or 0))
        for node in image_selector(doc)
        if bounds.upper < (position := dom.get_index(node)) < bounds.lower
    ]

    images = list(
        parse_image_nodes(
            image_nodes=image_nodes,
            caption_selector=caption_selector,
            alt_selector=alt_selector,
            author_selector=author_selector,
            domain=domain,
            size_pattern=size_pattern,
        )
    )

    return images



import time
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Iterable, Iterator, List, Optional, Protocol
from urllib.parse import urlparse

import chardet
import lxml.html
import requests
import validators
from fastwarc import ArchiveIterator, WarcRecord, WarcRecordType
from lxml.cssselect import CSSSelector
from lxml.etree import XPath
from requests import ConnectionError, HTTPError, ReadTimeout

from fundus.logging import create_logger
from fundus.publishers.base_objects import Publisher, Robots
from fundus.scraping.delay import Delay
from fundus.scraping.filter import URLFilter
from fundus.scraping.session import _default_header, session_handler
from fundus.scraping.url import URLSource
from fundus.utils.events import __EVENTS__

__all__ = [
    "HTML",
    "SourceInfo",
    "WarcSourceInfo",
    "WebSourceInfo",
    "HTMLSource",
    "WebSource",
    "CCNewsSource",
]

logger = create_logger(__name__)

# unfortunately lxml does not support case-insensitive CSSSelectors
_content_type_selector = CSSSelector("meta[http-equiv='Content-Type'], meta[http-equiv='content-type']")
_charset_selector = XPath("//meta[@charset]/@charset | //meta[@charSet]/@charSet")


def _detect_charset_from_response(response: requests.Response) -> str:
    """Detects HTML encoding based on meta tag <http-equiv='Content-Type'>

    Args:
        response: Response to detect encoding for

    Returns:
        str: detected encoding or response.apparent_encoding
    """
    # see https://github.com/flairNLP/fundus/issues/446
    # use response fallback to decode HTML in a first guess
    guessed_text = response.content.decode(response.encoding or "utf-8", errors="replace")
    document = lxml.html.document_fromstring(guessed_text)
    if (content_type_nodes := _content_type_selector(document)) and len(content_type_nodes) == 1:
        content_type_node = content_type_nodes.pop()
        for field in content_type_node.attrib.get("content", "").split(";"):
            if "charset" in field:
                charset = field.replace("charset=", "").strip()
                return charset
    elif charset := _charset_selector(document):
        return str(charset.pop())
    return response.apparent_encoding


@dataclass(frozen=True)
class HTML:
    requested_url: str
    responded_url: str
    content: str
    crawl_date: datetime
    source_info: "SourceInfo"


@dataclass(frozen=True)
class SourceInfo:
    publisher: str


@dataclass(frozen=True)
class WarcSourceInfo(SourceInfo):
    warc_path: str
    warc_headers: Dict[str, str]
    http_headers: Dict[str, str]


@dataclass(frozen=True)
class WebSourceInfo(SourceInfo):
    type: str
    url: str


class HTMLSource(Protocol):
    def fetch(self, url_filter: Optional[URLFilter] = None) -> Iterator[HTML]:
        ...


class WebSource:
    def __init__(
        self,
        url_source: Iterable[str],
        publisher: Publisher,
        url_filter: Optional[URLFilter] = None,
        request_header: Optional[Dict[str, str]] = None,
        query_parameters: Optional[Dict[str, str]] = None,
        delay: Optional[Delay] = None,
        ignore_robots: bool = False,
        ignore_crawl_delay: bool = False,
    ):
        self.url_source = url_source
        self.publisher = publisher
        self.url_filter = url_filter
        self.request_header = request_header or _default_header
        self.query_parameters = query_parameters or {}
        if isinstance(url_source, URLSource):
            url_source.set_header(self.request_header)

        self.delay = delay

        # register default events
        __EVENTS__.register_event("stop")

        # parse robots:
        self.robots: Optional[Robots] = None
        if not ignore_robots:
            self.robots = self.publisher.robots
            if not self.robots.ready:
                self.publisher.robots.read(headers=self.request_header)

            if not ignore_crawl_delay:
                if robots_delay := self.robots.crawl_delay(self.request_header.get("user-agent") or "*"):
                    logger.debug(
                        f"Found crawl-delay of {robots_delay} seconds in robots.txt for {self.publisher.name}. "
                        f"Overwriting existing delay."
                    )
                    self.delay = lambda: robots_delay

    @property
    def _is_stopped(self):
        return __EVENTS__.is_event_set("stop")

    @staticmethod
    def sleep(s: float):
        __EVENTS__.get("stop").wait(s)

    def fetch(self, url_filter: Optional[URLFilter] = None) -> Iterator[HTML]:
        combined_filters: List[URLFilter] = ([self.url_filter] if self.url_filter else []) + (
            [url_filter] if url_filter else []
        )

        timestamp = time.time() + self.delay() if self.delay is not None else time.time()

        def filter_url(u: str) -> bool:
            return any(f(u) for f in combined_filters)

        url_iterator = iter(self.url_source)

        while not self._is_stopped:
            if (url := next(url_iterator, None)) is None:
                return

            if not validators.url(url):
                logger.debug(f"Skipped requested URL {url!r} because the URL is malformed")
                continue

            if filter_url(url):
                logger.debug(f"Skipped requested URL {url!r} because of URL filter")
                continue

            if not (self.robots is None or self.robots.can_fetch(self.request_header.get("user-agent") or "*", url)):
                logger.debug(f"Skipped requested URL {url!r} because of robots.txt")
                continue

            session = session_handler.get_session()
            for key, value in self.query_parameters.items():
                if "?" in url:
                    url += "&" + key + "=" + value
                else:
                    url += "?" + key + "=" + value

            if self.delay:
                # Instead of using time.sleep, we use a custom sleep function that waits
                # for the "stop" event. This ensures that sleep does not block the shutdown process.
                self.sleep(max(0.0, self.delay() - time.time() + timestamp))
                timestamp = time.time()

            try:
                response = session.get_with_interrupt(url, headers=self.request_header)

            except (HTTPError, ConnectionError, ReadTimeout) as error:
                logger.warning(f"Skipped requested URL {url!r} because of {error!r}")
                if isinstance(error, HTTPError) and error.response.status_code >= 500:
                    logger.warning(f"Skipped {self.publisher.name!r} due to server errors: {error!r}")
                continue

            except Exception as error:
                logger.error(f"Warning! Skipped requested URL {url!r} because of an unexpected error {error!r}")
                continue

            else:
                if "charset" not in response.headers["content-type"]:
                    # That's actually the only place requests checks to detect encoding, so if charset
                    # is not set, requests falls back to default encodings (latin-1/utf-8)
                    logger.debug(f"Detect encoding from response for URL {str(response.url)!r}")
                    response.encoding = _detect_charset_from_response(response)

                if filter_url(str(response.url)):
                    logger.debug(f"Skipped responded URL {str(response.url)!r} because of URL filter")
                    continue
                html = response.text

                if response.history:
                    logger.info(f"Got redirected {len(response.history)} time(s) from {url!r} -> {response.url!r}")

                source_info = (
                    WebSourceInfo(self.publisher.name, type(self.url_source).__name__, self.url_source.url)
                    if isinstance(self.url_source, URLSource)
                    else SourceInfo(self.publisher.name)
                )

                yield HTML(
                    requested_url=url,
                    responded_url=str(response.url),
                    content=html,
                    crawl_date=datetime.now(),
                    source_info=source_info,
                )


class CCNewsSource:
    def __init__(self, *publishers: Publisher, warc_path: str, headers: Optional[Dict[str, str]] = None):
        self.publishers = publishers
        self.warc_path = warc_path
        self.headers = headers or _default_header
        self._publisher_mapping: Dict[str, Publisher] = {
            urlparse(publisher.domain).netloc: publisher for publisher in self.publishers
        }

    def fetch(self, url_filter: Optional[URLFilter] = None) -> Iterator[HTML]:
        def extract_content(record: WarcRecord) -> Optional[str]:
            warc_body: bytes = record.reader.read()

            try:
                return str(warc_body, encoding=record.http_charset)  # type: ignore[arg-type]
            except (UnicodeDecodeError, TypeError):
                encoding: Optional[str] = chardet.detect(warc_body)["encoding"]

                if encoding is not None:
                    logger.debug(
                        f"Trying to decode record {record.record_id!r} from {target_url!r} "
                        f"using detected encoding {encoding}."
                    )

                    try:
                        return str(warc_body, encoding=encoding)
                    except UnicodeDecodeError:
                        logger.warning(
                            f"Couldn't decode record {record.record_id!r} from {target_url!r} with "
                            f"original charset {record.http_charset!r} using detected charset {encoding!r}."
                        )
                else:
                    logger.warning(
                        f"Couldn't detect charset for record {record.record_id!r} from {target_url!r} "
                        f"with invalid original charset {record.http_charset!r}."
                    )

            return None

        with requests.Session() as session:
            response = session.get(self.warc_path, stream=True, headers=self.headers)
            response.raise_for_status()

            for warc_record in ArchiveIterator(response.raw, record_types=WarcRecordType.response, verify_digests=True):
                if not warc_record.record_date:
                    continue

                target_url = str(warc_record.headers["WARC-Target-URI"])

                if url_filter is not None and url_filter(target_url):
                    logger.debug(f"Skipped WARC record with target URI {target_url!r} because of URL filter")
                    continue

                publisher_domain: str = urlparse(target_url).netloc

                if publisher_domain not in self._publisher_mapping:
                    continue

                publisher = self._publisher_mapping[publisher_domain]

                if publisher.url_filter is not None and publisher.url_filter(target_url):
                    logger.debug(
                        f"Skipped WARC record with target URI {target_url!r} because of "
                        f"publisher specific URL filter"
                    )
                    continue

                if (content := extract_content(warc_record)) is None:
                    continue

                yield HTML(
                    requested_url=target_url,
                    responded_url=target_url,
                    content=content,
                    crawl_date=warc_record.record_date,
                    source_info=WarcSourceInfo(
                        publisher=publisher.name,
                        warc_path=self.warc_path,
                        warc_headers=dict(warc_record.headers),
                        http_headers=dict(warc_record.http_headers or {}),
                    ),
                )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_date_parsing,
    generic_nodes_to_text,
    generic_topic_parsing,
    image_extraction,
)


class NikkeiParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = XPath(
            "//section[@class='container_campx13']//*[self::div and @class='container_c1tzahnc' and position()=1]"
        )
        _paragraph_selector = CSSSelector("section[data-track-article-content] > p")
        _subheadline_selector = CSSSelector("section[data-track-article-content] > div > h2")

        _topic_selector = XPath("//article //header //div[contains(@class, 'topicLink')]")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def topics(self) -> List[str]:
            if topic_nodes := self._topic_selector(self.precomputed.doc):
                return generic_topic_parsing(generic_nodes_to_text(topic_nodes), "・")
            return []

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                lower_boundary_selector=CSSSelector("p.title_thchiij"),
                # https://regex101.com/r/qjEM41/1
                author_selector=re.compile(r"=(?P<credits>[^=]*?)\s*$"),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.publishers.za.daily_maverick import DailyMaverickParser
from fundus.publishers.za.times_live import TimesLiveParser
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import NewsMap, Sitemap


class ZA(metaclass=PublisherGroup):
    default_language = "en"

    DailyMaverick = Publisher(
        name="Daily Maverick",
        domain="https://www.dailymaverick.co.za/",
        parser=DailyMaverickParser,
        sources=[
            NewsMap("https://www.dailymaverick.co.za/news-sitemap.xml"),
            Sitemap(
                "https://www.dailymaverick.co.za/sitemap_index.xml",
                sitemap_filter=inverse(regex_filter("article-sitemap")),
            ),
        ],
    )

    TimesLive = Publisher(
        name="Times Live",
        domain="https://www.timeslive.co.za/",
        parser=TimesLiveParser,
        sources=[
            NewsMap("https://www.timeslive.co.za/sitemap/google-news/times-live/news/"),
            NewsMap("https://www.timeslive.co.za/sitemap/google-news/times-live/politics/"),
            NewsMap("https://www.timeslive.co.za/sitemap/google-news/times-live/sport/"),
            NewsMap("https://www.timeslive.co.za/sitemap/google-news/times-live/lifestyle/"),
            NewsMap("https://www.timeslive.co.za/sitemap/google-news/sunday-times/news/"),
            NewsMap("https://www.timeslive.co.za/sitemap/google-news/sunday-times/business/"),
            NewsMap("https://www.timeslive.co.za/sitemap/google-news/sunday-times-daily/news/"),
        ],
    )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.url import Sitemap

from .sermitsiaq import SermitsiaqParser


class GL(metaclass=PublisherGroup):
    default_language = "kl"

    Sermitsiaq = Publisher(
        name="Sermitsiaq",
        domain="https://www.sermitsiaq.ag/",
        parser=SermitsiaqParser,
        sources=[Sitemap("https://www.sermitsiaq.ag/sitemap.xml")],
    )



import datetime
import re
from typing import List, Match, Optional, Pattern

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    image_extraction,
)


class FocusParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = CSSSelector("div.textBlock > p , div[data-qa-article-content-text] > p")
        _summary_selector = CSSSelector("div.leadIn > p, div.Article-Description ")
        _subheadline_selector = CSSSelector("div.textBlock > h2, div[data-qa-article-content-text] > h2")
        _snippet_selector = XPath(
            'string(//script[@type="text/javascript"][contains(text(), "window.bf__bfa_metadata")])'
        )

        # regex patterns
        _author_substitution_pattern: Pattern[str] = re.compile(
            r"Von FOCUS-online-(Redakteur|Autorin|Reporter|Redakteurin|Gastautor)\s"
        )
        _topic_pattern: Pattern[str] = re.compile(r'"keywords":\[{(.*?)}\]')
        _topic_name_pattern: Pattern[str] = re.compile(r'"name":"(.*?)"', flags=re.MULTILINE)

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            author_names = generic_author_parsing(self.precomputed.ld.bf_search("author"))
            for i, name in enumerate(author_names):
                author_names[i] = re.sub(self._author_substitution_pattern, "", name)
            return author_names

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute(deprecated=datetime.date(2024, 6, 1))
        def topics(self) -> List[str]:
            snippet = self._snippet_selector(self.precomputed.doc)
            if not snippet:
                return []

            match: Optional[Match[str]] = re.search(self._topic_pattern, snippet)
            if not match:
                return []
            topic_names: List[str] = re.findall(self._topic_name_pattern, match.group(1))

            return topic_names

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//div[@class='image clearfix']//img|//figure//img"),
                caption_selector=XPath(
                    "./ancestor::div[@class='image clearfix']//span[@class='caption']|"
                    "./ancestor::figure//span[@class='Image-Caption']"
                ),
                author_selector=XPath(
                    "./ancestor::div[@class='image clearfix']//span[@class='source']|"
                    "./ancestor::figure//span[@class='Image-Credit']"
                ),
                lower_boundary_selector=XPath("//footer"),
            )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    apply_substitution_pattern_over_list,
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class AsahiShimbunParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = CSSSelector("div.nfyQp > div.bv2Sj > p")
        _paragraph_selector = CSSSelector("div.nfyQp > p")
        _subtitle_selector = CSSSelector("div.nfyQp > h2")

        topic_bloat_pattern = re.compile(r"朝日新聞デジタル|朝日新聞|ニュース|新聞|その他・話題")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subtitle_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("TITLE")

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.meta.get("article:published_time"))

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def topics(self) -> List[str]:
            return apply_substitution_pattern_over_list(
                generic_topic_parsing(self.precomputed.meta.get("keywords")), self.topic_bloat_pattern
            )

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                author_selector=re.compile(r"、(?P<credits>[^、]*?)撮影"),
                relative_urls=True,
            )



import datetime
import re
from typing import List, Optional, Pattern

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    apply_substitution_pattern_over_list,
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class SeznamZpravyParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath(
            "//div[contains(@class,'mol-rich-content--for-article')]/div[contains(@class,'speakable')]/p"
        )
        _summary_selector = XPath("//div/p[contains(@class, 'speakable') and @*[contains(., 'ogm-article-perex')]]")
        _subheadline_selector = XPath("//div[contains(@class,'mol-rich-content--for-article')]/h2")
        _author_substitution_pattern: Pattern[str] = re.compile(r"Seznam Zprávy")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.ld.bf_search("keywords"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def authors(self) -> List[str]:
            return apply_substitution_pattern_over_list(
                generic_author_parsing(self.precomputed.ld.bf_search("author")),
                pattern=self._author_substitution_pattern,
                replacement="",
            )

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//figure//img[not(ancestor::div[contains(@class, 'mol-post-card__body')])]"),
                author_selector=XPath("./ancestor::figure//span[@*[contains(., 'atm-media-item-image-caption')]]"),
                relative_urls=True,
            )



from __future__ import annotations

import contextlib
import gzip
import json
import logging.config
import multiprocessing
import os
import random
import re
import time
import traceback
from abc import ABC, abstractmethod
from collections import defaultdict
from datetime import datetime
from functools import lru_cache, partial, wraps
from multiprocessing import Manager
from multiprocessing.context import TimeoutError
from multiprocessing.managers import BaseManager
from multiprocessing.pool import MapResult, Pool, ThreadPool
from pathlib import Path
from queue import Empty, Queue
from threading import current_thread
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Literal,
    Optional,
    Pattern,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import dill
import fastwarc.stream_io
import more_itertools
import requests
import urllib3.exceptions
from dateutil.rrule import MONTHLY, rrule
from more_itertools import roundrobin
from tqdm import tqdm
from typing_extensions import ParamSpec, TypeAlias

from fundus.logging import create_logger, get_current_config
from fundus.parser.data import remove_query_parameters_from_url
from fundus.publishers.base_objects import FilteredPublisher, Publisher, PublisherGroup
from fundus.scraping.article import Article
from fundus.scraping.delay import Delay
from fundus.scraping.filter import ExtractionFilter, Requires, RequiresAll, URLFilter
from fundus.scraping.html import CCNewsSource
from fundus.scraping.scraper import CCNewsScraper, WebScraper
from fundus.scraping.session import session_handler
from fundus.scraping.url import URLSource
from fundus.utils.events import __EVENTS__
from fundus.utils.timeout import Timeout

logger = create_logger(__name__)

_T = TypeVar("_T")
_P = ParamSpec("_P")

PublisherType: TypeAlias = Union[Publisher, PublisherGroup]


class RemoteException(Exception):
    pass


class TQDMManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register("_tqdm", tqdm)

    def tqdm(self, *args, **kwargs) -> tqdm:
        return getattr(self, "_tqdm")(*args, **kwargs)


@contextlib.contextmanager
def get_proxy_tqdm(*args, **kwargs) -> tqdm:
    """
    This functions returns a proxy to a tqdm instance. Init args are the same as for any other tqdm instance.
    :param args: tqdm args
    :param kwargs: tqdm kwargs
    :return: a self-managed, proxied tqdm instance
    """
    manager = TQDMManager()
    try:
        manager.start()
        yield manager.tqdm(*args, **kwargs)
    finally:
        manager.shutdown()


# noinspection PyPep8Naming
class dill_wrapper(Generic[_P, _T]):
    def __init__(self, target: Callable[_P, _T]):
        """Wraps function in dill serialization.

        This is in order to use unpickable functions within multiprocessing.

        Args:
            target: The function to wrap.
        """
        self._serialized_target: bytes = dill.dumps(target)

    @lru_cache
    def _deserialize(self) -> Callable[_P, _T]:
        return cast(Callable[_P, _T], dill.loads(self._serialized_target))

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _T:
        return self._deserialize()(*args, **kwargs)


def get_execution_context():
    """
    Determines whether the current execution context is in a thread or process.
    Returns:
        context (str): "thread" or "process"
        ident (int): Thread ID or Process ID
    """
    if multiprocessing.current_process().name != "MainProcess":
        process = multiprocessing.current_process()
        return process.name, process.ident
    else:
        thread = current_thread()
        return thread.name, thread.ident


def queue_wrapper(queue: Queue[Union[_T, Exception]], target: Callable[_P, Iterator[_T]]) -> Callable[_P, None]:
    """Wraps the target callable to add its results to the queue instead of returning them directly.

    Args:
        queue: The buffer queue.
        target: A target callable.

    Returns:
        (Callable[_P, None]) The wrapped target.
    """

    @wraps(target)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> None:
        try:
            for obj in target(*args, **kwargs):
                queue.put(obj)
        except Exception as err:
            tb_str = "".join(traceback.TracebackException.from_exception(err).format())
            context, ident = get_execution_context()
            queue.put(
                RemoteException(
                    f"There was a(n) {type(err).__name__!r} occurring in {context} with ident {ident}\n{tb_str}"
                )
            )

            logger.debug(f"Encountered remote exception: {err!r}")
            # prevents a race condition where the ThreadPool shuts down before the exception is pulled from the queue
            time.sleep(0.2)

    return wrapper


def pool_queue_iter(handle: MapResult[Any], queue: Queue[Union[_T, Exception]]) -> Iterator[_T]:
    """Utility function to iterate exhaustively over a pool queue.

    The underlying iterator of this function repeatedly exhausts the given queue.
    Then, if the queue is empty only if all the pool's jobs have finished, the iterator reruns.
    Otherwise, it waits for the queue to be populated with the next result from the pool.

    Args:
        handle:  A handle of the MappedResult of the underling multiprocessing pool.
        queue: The pool queue.

    Returns:
        Iterator[_T]: The iterator over the queue as it is populated.
    """
    while True:
        try:
            if isinstance(nxt := queue.get_nowait(), Exception):
                raise Exception("There was an exception occurring in a remote thread/process") from nxt
            yield nxt
        except Empty:
            try:
                handle.get(timeout=0.1)
            except TimeoutError:
                if __EVENTS__.is_event_set("stop"):
                    __EVENTS__.clear_event("stop")
                    break
                continue
            return


def random_sleep(func: Callable[_P, _T], between: Tuple[float, float]) -> Callable[_P, _T]:
    @wraps(func)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        time.sleep(random.uniform(*between))
        return func(*args, **kwargs)

    return wrapper


class CrawlerBase(ABC):
    def __init__(self, *publishers: PublisherType):
        self.publishers: List[Union[Publisher, FilteredPublisher]] = list(set(more_itertools.collapse(publishers)))
        if not self.publishers:
            raise ValueError("param <publishers> of <Crawler.__init__> must include at least one publisher.")

        __EVENTS__.alias("main-thread")
        __EVENTS__.register_event("stop")

    @abstractmethod
    def _build_article_iterator(
        self,
        publishers: Tuple[Publisher, ...],
        error_handling: Literal["suppress", "catch", "raise"],
        extraction_filter: Optional[ExtractionFilter],
        url_filter: Optional[URLFilter],
        language_filter: Optional[List[str]],
    ) -> Iterator[Article]:
        raise NotImplementedError

    def crawl(
        self,
        max_articles: Optional[int] = None,
        max_articles_per_publisher: Optional[int] = None,
        timeout: Optional[int] = None,
        error_handling: Literal["suppress", "catch", "raise"] = "suppress",
        only_complete: Union[bool, ExtractionFilter] = Requires("title", "body", "publishing_date"),
        url_filter: Optional[URLFilter] = None,
        language_filter: Optional[List[str]] = None,
        only_unique: bool = True,
        save_to_file: Union[None, str, Path] = None,
    ) -> Iterator[Article]:
        """Yields articles from initialized scrapers

        Args:
            max_articles (Optional[int]): Number of articles to crawl. If there are fewer articles
                than max_articles the Iterator will stop before max_articles. If None, all retrievable
                articles are returned. Defaults to None.
            max_articles_per_publisher: Specify the number of articles to crawl per publisher.
                Disables <max_articles>. Defaults to None.
            timeout (Optional[int]): timeout (Optional[int]): Specifies the duration in seconds the crawler
                will wait without receiving any articles before stopping. If set <= 0, or if not provided,
                the crawler will run until all sources are exhausted. Defaults to None.
            error_handling (Literal["suppress", "catch", "raise"]): Define how to handle errors
                encountered during extraction. If set to "suppress", all errors will be skipped, either
                with None values for respective attributes in the extraction or by skipping entire articles.
                If set to "catch", errors will be caught as attribute values or, if an entire article fails,
                through Article.exception. If set to "raise" all errors encountered during extraction will
                be raised. Defaults to "suppress".
            only_complete (Union[bool, ExtractionFilter]): Set a callable satisfying the ExtractionFilter
                protocol as an extraction filter or use a boolean. If False, all articles will be yielded,
                if True, only those with all attributes extracted. Defaults to ExtractionFilter letting
                through all articles with at least title, body, and publishing_date set.
            url_filter (Optional[URLFilter]): A callable object satisfying the URLFilter protocol to skip
                URLs before download. This filter applies on both requested and responded URL. Defaults to None.
            language_filter (Optional[List[str]]): A set of language codes to filter the articles by. If set,
                articles of different languages will be skipped and not counted towards the article count. Defaults
                to None.
            only_unique (bool): If set to True, articles yielded will be unique on the responded URL.
                Always returns the first encountered article. Defaults to True.
            save_to_file (Union[None, str, Path]): If set, the crawled articles will be collected saved to the
                specified file as a JSON list.

        Returns:
            Iterator[Article]: An iterator yielding objects of type Article.
        """

        if max_articles == 0:
            return

        max_articles = max_articles or -1
        timeout = timeout or -1

        if max_articles_per_publisher:
            if timeout < 120:
                print(
                    "It is recommended to set a minimum <timeout> of 120 seconds when using "
                    "max_articles_per_publisher."
                )
            max_articles = -1

        def build_extraction_filter() -> Optional[ExtractionFilter]:
            if isinstance(only_complete, bool):
                return None if only_complete is False else RequiresAll()
            else:
                return only_complete

        response_cache: Set[str] = set()

        extraction_filter = build_extraction_filter()
        fitting_publishers: List[Union[Publisher, FilteredPublisher]] = []

        if isinstance(extraction_filter, Requires):
            for publisher in self.publishers:
                supported_attributes = set(
                    more_itertools.flatten(
                        collection.names for collection in publisher.parser.attribute_mapping.values()
                    )
                )
                if missing_attributes := extraction_filter.required_attributes - supported_attributes:
                    logger.warning(
                        f"The required attribute(s) `{', '.join(missing_attributes)}` "
                        f"is(are) not supported by {publisher.name}. Skipping publisher"
                    )
                elif language_filter and not publisher.supports(languages=language_filter):
                    logger.warning(
                        f"None of the required language(s) `{', '.join(language_filter)}` "
                        f"is(are) supported by {publisher.name}. Skipping publisher"
                    )
                else:
                    fitting_publishers.append(publisher)

            if not fitting_publishers:
                logger.error(
                    f"Could not find any fitting publishers for required attributes  "
                    f"`{', '.join(extraction_filter.required_attributes)}`"
                )
                return
        else:
            fitting_publishers = self.publishers

        # check if there are filtered publishers and if so, adopt their language restrictions
        publisher_language_filter = set()
        for publisher in fitting_publishers:
            if isinstance(publisher, FilteredPublisher):
                publisher_language_filter.update(publisher.language_filter)

        if language_filter and publisher_language_filter:
            language_filter = list(set(language_filter).union(publisher_language_filter))
            logger.info(
                f"Publisher language filter: {publisher_language_filter} will be added to the given language filter: "
                f"{language_filter}. "
            )
        elif publisher_language_filter:
            language_filter = list(publisher_language_filter)
            logger.info(f"Publisher language filter: {publisher_language_filter} will be used as the language filter. ")

        article_count: Dict[str, int] = defaultdict(int)
        crawled_articles: Dict[str, List[Article]] = defaultdict(list)

        # Unfortunately we relly on this little workaround here to terminate the 'Pool' used within
        # the 'CCNewsCrawler'. The 'Timeout' contextmanager utilizes '_thread.interrupt_main',
        # throwing a KeyboardInterrupt in the main thread after <time> seconds. My guess (MaxDall)
        # is, that within 'queue_wrapper's 'handle.get(timeout=0.1)', the main thread cannot be
        # interrupted via a KeyboardInterrupt. The workaround is to have a modul global event
        # that can be set within the 'Timeout' thread using a callback.
        # With Python 3.10 we can pass a signum to '_thread.interrupt_main', maybe that's the way to go.
        callback: Optional[Callable[[], None]]
        if isinstance(self, CCNewsCrawler) and self.processes > 0:

            def callback() -> None:
                __EVENTS__.set_event("stop", "main-thread")

        else:
            callback = None

        try:
            with Timeout(seconds=timeout, silent=True, callback=callback, disable=timeout <= 0) as timer:
                for article in self._build_article_iterator(
                    tuple(fitting_publishers), error_handling, build_extraction_filter(), url_filter, language_filter
                ):
                    if max_articles_per_publisher and article_count[article.publisher] == max_articles_per_publisher:
                        if isinstance(self, Crawler) and not __EVENTS__.is_event_set("stop", article.publisher):
                            __EVENTS__.set_event("stop", article.publisher)
                        if sum(article_count.values()) == len(self.publishers) * max_articles_per_publisher:
                            break
                        continue
                    timer.reset()
                    url_without_query_parameters = remove_query_parameters_from_url(article.html.responded_url)
                    if not only_unique or url_without_query_parameters not in response_cache:
                        response_cache.add(url_without_query_parameters)
                        article_count[article.publisher] += 1
                        if save_to_file:
                            crawled_articles[article.publisher].append(article)
                        yield article
                    if sum(article_count.values()) == max_articles:
                        break
        finally:
            session_handler.close_current_session()
            if save_to_file is not None:
                if isinstance(save_to_file, str):
                    save_to_file = Path(save_to_file)
                save_to_file.parent.mkdir(parents=True, exist_ok=True)
                with open(save_to_file, "w", encoding="utf-8") as file:
                    logger.info(f"Writing crawled articles to {save_to_file!r}")
                    file.write(
                        json.dumps(crawled_articles, default=lambda o: o.to_json(), ensure_ascii=False, indent=4)
                    )


class Crawler(CrawlerBase):
    def __init__(
        self,
        *publishers: PublisherType,
        restrict_sources_to: Optional[List[Type[URLSource]]] = None,
        ignore_deprecated: bool = False,
        delay: Optional[Union[float, Delay]] = 1.0,
        threading: bool = True,
        ignore_robots: bool = False,
        ignore_crawl_delay: bool = False,
    ):
        """Fundus base class for crawling articles from the web.

        Examples:
            >>> from fundus import PublisherCollection, Crawler
            >>> crawler = Crawler(*PublisherCollection)
            >>> # Crawler(PublisherCollection.us) to crawl only american news
            >>> for article in crawler.crawl():
            >>>     print(article)

        Args:
            *publishers (Union[Publisher, PublisherGroup]): The publishers to crawl.
            restrict_sources_to (Optional[List[Type[URLSource]]]): Lets you restrict sources defined in the publisher
                specs. If set, only articles from given source types will be yielded.
            ignore_deprecated (bool): If set to True, Publishers marked as deprecated will be skipped.
                Defaults to False.
            delay (Optional[Union[float, Delay]]): Set a delay time in seconds to be used between article
                downloads. You can set a delay directly using float or any callable satisfying the Delay
                protocol. If set to None, no delay will be used between batches. See Delay for more
                information. Defaults to None.
            threading (bool): If True, the crawler will use a dedicated thread per publisher, if set to False,
                the crawler will use a single thread for all publishers and load articles successively. This will
                greatly influence performance, and it is highly recommended to use a threaded crawler.
                Defaults to True.
            ignore_robots (bool): Determines whether to bypass the consideration of the robots.txt file when
                filtering URLs from publishers. If set to True, the URLs will not be filtered based on the
                robots.txt file. Defaults to False.
            ignore_crawl_delay (bool): Determines whether to ignore a crawl delay given by a publisher.
                If set to False, this will overwrite <delay>. If ignore_robots is set to True, the crawl delay
                will also be ignored.
        """

        def filter_publishers(publisher: Publisher) -> bool:
            if publisher.deprecated and ignore_deprecated:
                logger.warning(f"Skipping deprecated publisher: {publisher.name}")
                return False
            return True

        fitting_publishers = list(filter(filter_publishers, more_itertools.collapse(publishers)))
        if not fitting_publishers:
            raise ValueError(
                f"All given publishers are deprecated. Either set <ignore_deprecated> to `False` or "
                f"include at least one publisher that isn't deprecated."
            )

        super().__init__(*fitting_publishers)

        self.restrict_sources_to = restrict_sources_to
        self.delay = delay
        self.threading = threading
        self.ignore_robots = ignore_robots
        self.ignore_crawl_delay = ignore_crawl_delay

    def _fetch_articles(
        self,
        publisher: Publisher,
        error_handling: Literal["suppress", "catch", "raise"],
        extraction_filter: Optional[ExtractionFilter] = None,
        url_filter: Optional[URLFilter] = None,
        language_filter: Optional[List[str]] = None,
    ) -> Iterator[Article]:
        def build_delay() -> Optional[Delay]:
            if isinstance(self.delay, float):
                delay = self.delay

                def constant_delay() -> float:
                    return delay

                return constant_delay

            elif isinstance(self.delay, Delay):
                return self.delay

            else:
                raise TypeError("param <delay> of <Crawler.__init__>")

        scraper = WebScraper(
            publisher,
            self.restrict_sources_to,
            build_delay(),
            ignore_robots=self.ignore_robots,
            ignore_crawl_delay=self.ignore_crawl_delay,
        )
        if not scraper.sources and self.restrict_sources_to:
            logger.warning(
                f"No sources of type {[source_type.__name__ for source_type in self.restrict_sources_to]} found for publisher {publisher.name}. "
                f"Skipping publisher."
            )
            return
        yield from scraper.scrape(error_handling, extraction_filter, url_filter, language_filter)

    @staticmethod
    def _single_crawl(
        publishers: Tuple[Publisher, ...], article_task: Callable[[Publisher], Iterator[Article]]
    ) -> Iterator[Article]:
        article_iterators = [article_task(publisher) for publisher in publishers]
        yield from roundrobin(*article_iterators)

    def _threaded_crawl(
        self, publishers: Tuple[Publisher, ...], article_task: Callable[[Publisher], Iterator[Article]]
    ) -> Iterator[Article]:
        result_queue: Queue[Union[Article, Exception]] = Queue(len(publishers))
        wrapped_article_task = queue_wrapper(result_queue, article_task)
        pool = ThreadPool(processes=len(publishers) or None)
        try:
            with session_handler.context(
                POOL_CONNECTIONS=len(publishers),
            ):
                yield from pool_queue_iter(pool.map_async(wrapped_article_task, publishers), result_queue)
        finally:
            logger.debug(f"Shutting down {type(self).__name__!r} ...")
            __EVENTS__.set_for_all("stop")
            pool.close()
            pool.join()
            __EVENTS__.clear_for_all("stop")
            logger.debug("Shutdown done")

    def _build_article_iterator(
        self,
        publishers: Tuple[Publisher, ...],
        error_handling: Literal["suppress", "catch", "raise"],
        extraction_filter: Optional[ExtractionFilter],
        url_filter: Optional[URLFilter],
        language_filter: Optional[List[str]],
    ) -> Iterator[Article]:
        article_task = partial(
            self._fetch_articles,
            error_handling=error_handling,
            extraction_filter=extraction_filter,
            url_filter=url_filter,
            language_filter=language_filter,
        )

        if self.threading:
            yield from self._threaded_crawl(publishers, article_task)
        else:
            yield from self._single_crawl(publishers, article_task)


class CCNewsCrawler(CrawlerBase):
    def __init__(
        self,
        *publishers: PublisherType,
        start: datetime = datetime(2016, 8, 1),
        end: datetime = datetime.now(),
        processes: int = -1,
        retries: int = 3,
        disable_tqdm: bool = False,
        server_address: str = "https://data.commoncrawl.org/",
    ):
        """Initializes a crawler for the CC-NEWS dataset.

        The crawler crawls the CC-NEWS dataset from <end> to <start>.

        Args:
            *publishers: The publishers to crawl.
            start: The date to start crawling from. Refers to the date the WARC record was added to CC-NEWS,
                not when it was published. Defaults to 2016/8/1.
            end: The date to end crawling. Refers to the date the WARC record was added to CC-NEWS, not when
                it was published. Defaults to datetime.now().
            processes: Number of additional process to use for crawling.
                If -1, the number of processes is set to `os.cpu_count()`.
                If `os.cpu_count()` is not available, the number of processes is set to 0.
                If 0, only the main process is used. Defaults to -1.
            retries: The number of times to retry crawling a WARC record when a connection error occurs. Between
                retries, the crawler sleeps for <current-try> * 30 seconds. Defaults to 3.
            disable_tqdm: Disable the usage of tqdm within the crawler. Defaults to False.
            server_address: The CC-NEWS dataset server address. Defaults to 'https://data.commoncrawl.org/'.
        """

        super().__init__(*publishers)

        self.start = start
        self.end = end

        if processes < 0:
            print(
                f"{type(self).__name__} will automatically use all available cores: {os.cpu_count()}. "
                f"For optimal performance, we recommend manually setting the number of processes "
                f"using the <processes> parameter. A good rule of thumb is to allocate `one process per "
                f"200 Mbps of bandwidth`."
            )
            self.processes = os.cpu_count() or 0
        else:
            self.processes = processes

        self.retries = retries
        self.disable_tqdm = disable_tqdm
        self.server_address = server_address

    def _fetch_articles(
        self,
        warc_path: str,
        publishers: Tuple[Publisher, ...],
        error_handling: Literal["suppress", "catch", "raise"],
        extraction_filter: Optional[ExtractionFilter] = None,
        url_filter: Optional[URLFilter] = None,
        language_filter: Optional[List[str]] = None,
        bar: Optional[tqdm] = None,
    ) -> Iterator[Article]:
        retries: int = 0
        while True:
            source = CCNewsSource(*publishers, warc_path=warc_path)
            scraper = CCNewsScraper(source)
            try:
                yield from scraper.scrape(error_handling, extraction_filter, url_filter, language_filter)
            except (requests.HTTPError, fastwarc.stream_io.StreamError, urllib3.exceptions.HTTPError) as exception:
                if retries >= self.retries:
                    logger.error(f"Failed to load WARC file {warc_path!r} after {retries} retries")
                    break
                else:
                    retries += 1
                    sleep_time = (30 * retries) + random.uniform(-2, 2)
                    logger.warning(
                        f"Could not load WARC file {warc_path!r}. Retry after {sleep_time:.2f} seconds: {exception!r}"
                    )
                    time.sleep(sleep_time)
            else:
                break

        if bar is not None:
            bar.update()

    @staticmethod
    def _single_crawl(
        warc_paths: Tuple[str, ...], article_task: Callable[[str], Iterator[Article]]
    ) -> Iterator[Article]:
        for warc_path in warc_paths:
            yield from article_task(warc_path)

    def _parallel_crawl(
        self, warc_paths: Tuple[str, ...], article_task: Callable[[str], Iterator[Article]]
    ) -> Iterator[Article]:
        # because logging configurations are overwritten when using 'spawn' as start method,
        # we have to get current logging configurations and initialize them in the new process
        if multiprocessing.get_start_method() == "spawn":
            logging_config = get_current_config()
            initializer = partial(logging.config.dictConfig, config=logging_config)
        else:
            initializer = None

        # As one could think, because we're downloading a bunch of files, this task is IO-bound, but it is actually
        # process-bound. The reason is that we stream the data and process it on the fly rather than downloading all
        # files and processing them afterward. Therefore, we utilize multiprocessing here instead of multithreading.
        try:
            with Manager() as manager, Pool(
                processes=min(self.processes, len(warc_paths)),
                initializer=initializer,
            ) as pool:
                result_queue: Queue[Union[Article, Exception]] = manager.Queue(maxsize=1000)

                # Because multiprocessing.Pool does not support iterators as targets,
                # we wrap the article_task to write the articles to a queue instead of returning them directly.
                wrapped_article_task: Callable[[str], None] = queue_wrapper(result_queue, article_task)

                # To avoid 503 errors we spread tasks to not start all at once
                spread_article_task = random_sleep(wrapped_article_task, (0, 3))

                # To avoid restricting the article_task to use only pickleable objects, we serialize it using dill.
                serialized_article_task = dill_wrapper(spread_article_task)

                # Finally, we build an iterator around the queue, exhausting the queue until the pool is finished.
                yield from pool_queue_iter(pool.map_async(serialized_article_task, warc_paths), result_queue)
        finally:
            logger.debug(f"Shutting down {type(self).__name__!r} ...")
            logger.debug("Joining manager ...")
            manager.join()
            logger.debug("Joining pool ...")
            pool.join()
            logger.debug("Shutdown done")

    def _get_warc_paths(self) -> List[str]:
        # Date regex examples: https://regex101.com/r/yDX3G6/1
        date_pattern: Pattern[str] = re.compile(r"CC-NEWS-(?P<date>\d{14})-")

        if self.start >= self.end:
            raise ValueError("Start date has to be < end date.")

        if self.start < datetime(2016, 8, 1):
            raise ValueError("The default, and earliest possible, start date is 2016/08/01.")

        if self.end > datetime.now():
            raise ValueError("The specified end date is in the future. We don't want to give spoilers, do we?")

        date_sequence: List[datetime] = list(rrule(MONTHLY, dtstart=self.start, until=self.end))
        urls: List[str] = [
            f"{self.server_address}crawl-data/CC-NEWS/{date.strftime('%Y/%m')}/warc.paths.gz" for date in date_sequence
        ]

        with tqdm(total=len(urls), desc="Loading WARC Paths", leave=False, disable=self.disable_tqdm) as bar:

            def load_paths(url: str) -> List[str]:
                with requests.Session() as session:
                    paths = gzip.decompress(session.get(url).content).decode("utf-8").split()
                    bar.update()
                    return paths

            if self.processes == 0:
                nested_warc_paths = [load_paths(url) for url in urls]
            else:
                # use two threads per process, default two threads per core
                max_number_of_threads = self.processes * 2

                try:
                    with ThreadPool(processes=min(len(urls), max_number_of_threads)) as pool:
                        nested_warc_paths = pool.map(random_sleep(load_paths, (0, 3)), urls)
                finally:
                    pool.join()

        warc_paths: Iterator[str] = more_itertools.flatten(nested_warc_paths)

        start_strf = self.start.strftime("%Y%m%d%H%M%S")
        end_strf = self.end.strftime("%Y%m%d%H%M%S")

        def filter_warc_path_by_date(path: str) -> bool:
            match: Optional[re.Match[str]] = date_pattern.search(path)
            if match is None:
                raise AssertionError(f"Invalid WARC path {path!r}")
            return start_strf <= match["date"] <= end_strf

        return sorted(
            (f"{self.server_address}{warc_path}" for warc_path in filter(filter_warc_path_by_date, warc_paths)),
            reverse=True,
        )

    def _build_article_iterator(
        self,
        publishers: Tuple[Publisher, ...],
        error_handling: Literal["suppress", "catch", "raise"],
        extraction_filter: Optional[ExtractionFilter],
        url_filter: Optional[URLFilter],
        language_filter: Optional[List[str]],
        **kwargs,
    ) -> Iterator[Article]:
        warc_paths = tuple(self._get_warc_paths())

        with get_proxy_tqdm(total=len(warc_paths), desc="Process WARC files", disable=self.disable_tqdm) as bar:
            article_task = partial(
                self._fetch_articles,
                publishers=publishers,
                error_handling=error_handling,
                extraction_filter=extraction_filter,
                url_filter=url_filter,
                language_filter=language_filter,
                bar=bar,
            )

            if self.processes == 0:
                yield from self._single_crawl(warc_paths, article_task)
            else:
                yield from self._parallel_crawl(warc_paths, article_task)



from pickle import dumps, loads
from typing import Any, Dict, List

from lxml.etree import XPath, tostring

from fundus.parser.data import LinkedDataMapping

lds: List[Dict[str, Any]] = [
    {"@type": "Example1", "value": 1, "data": {"inner": "value", "nested": {"dict": True}}},
    {"@type": "Example2", "value": 2, "_:*@": "Howdy"},
    {"this": "should", "be": {"of": "type", "__UNKNOWN_TYPE__": True, "value": 3}},
]


class TestLinkedDataMapping:
    def test_constructor(self):
        LinkedDataMapping(lds)

    def test_xpath_search(self):
        ld = LinkedDataMapping(lds)
        assert ld.xpath_search(XPath("//value")) == ["1", "2", "3"]
        assert ld.xpath_search(XPath("//UNKNOWN_TYPE//value")) == ["3"]
        assert ld.xpath_search(XPath("//_U003AU002AU0040")) == ["Howdy"]
        assert ld.xpath_search(XPath("//dict")) == ["True"]
        assert ld.xpath_search(XPath("//Example2")) == [{"@type": "Example2", "value": "2", "_:*@": "Howdy"}]

    def test_pickle(self):
        ld = LinkedDataMapping(lds)
        ld.__as_xml__()
        ld_pickled = loads(dumps(ld))
        assert ld_pickled.__getattribute__("Example1") == ld.__getattribute__("Example1")
        assert ld_pickled.__getattribute__("Example2") == ld.__getattribute__("Example2")
        assert ld_pickled.__getattribute__("UNKNOWN_TYPE") == ld.__getattribute__("UNKNOWN_TYPE")
        assert tostring(ld_pickled.__getattribute__("_LinkedDataMapping__xml")) == tostring(
            ld.__getattribute__("_LinkedDataMapping__xml")
        )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.publishers.ca.cbc_news import CBCNewsParser
from fundus.publishers.ca.globe_and_mail import TheGlobeAndMailParser
from fundus.publishers.ca.national_post import NationalPostParser
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

# noinspection PyPep8Naming


class CA(metaclass=PublisherGroup):
    default_language = "en"

    CBCNews = Publisher(
        name="CBC News",
        domain="https://www.cbc.ca/",
        parser=CBCNewsParser,
        sources=[
            RSSFeed("https://www.cbc.ca/webfeed/rss/rss-topstories"),
            RSSFeed("https://www.cbc.ca/webfeed/rss/rss-world"),
            RSSFeed("https://www.cbc.ca/webfeed/rss/rss-canada"),
        ],
    )
    TheGlobeAndMail = Publisher(
        name="The Globe and Mail",
        domain="https://www.theglobeandmail.com",
        parser=TheGlobeAndMailParser,
        sources=[
            NewsMap("https://www.theglobeandmail.com/arc/outboundfeeds/news-sitemap-index/?outputType=xml"),
            NewsMap("https://www.theglobeandmail.com/arc/outboundfeeds/sitemap-index/?outputType=xml"),
        ],
    )

    NationalPost = Publisher(
        name="National Post",
        domain="https://nationalpost.com",
        parser=NationalPostParser,
        sources=[
            NewsMap("https://nationalpost.com/sitemap-news.xml"),
            Sitemap("https://nationalpost.com/sitemap-old.xml"),
            RSSFeed("https://nationalpost.com/feed"),
        ],
    )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    apply_substitution_pattern_over_list,
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
    normalize_whitespace,
)


class MainichiShimbunParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = CSSSelector("#articledetail-body > p")
        _subheadline_selector = CSSSelector("#articledetail-body > h2")

        _topic_bloat_pattern = re.compile("速報")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            if (title := self.precomputed.meta.get("title")) is not None:
                return normalize_whitespace(title)
            return None

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.meta.get("cXenseParse:author"))

        @attribute
        def topics(self) -> List[str]:
            return apply_substitution_pattern_over_list(
                generic_topic_parsing(self.precomputed.meta.get("keywords"), delimiter=[",", "・"]),
                self._topic_bloat_pattern,
            )

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//figure//img[not(ancestor::a[contains(@class,'articledetail-image-scale')])]"),
                upper_boundary_selector=CSSSelector("#main"),
                # https://regex101.com/r/awU0Rq/1
                author_selector=re.compile(r"(、|＝(?=.*?撮影$))(?P<credits>[^、]*?)(撮影)?\s*$"),
                relative_urls=True,
            )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
    parse_title_from_root,
)


class PeopleParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = CSSSelector("div.rm_txt_con > p")

        _author_selector = CSSSelector("div.edit")
        _author_pattern = re.compile(r"：(.*)\)")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(self.precomputed.doc, paragraph_selector=self._paragraph_selector)

        @attribute
        def title(self) -> Optional[str]:
            return parse_title_from_root(self.precomputed.doc)

        @attribute
        def authors(self) -> List[str]:
            if (author_nodes := self._author_selector(self.precomputed.doc)) and len(author_nodes) == 1:
                if match := re.search(self._author_pattern, author_nodes.pop().text_content()):
                    return generic_author_parsing(match.group(1), split_on=["、"])
            return []

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.meta.get("publishdate"))

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"), delimiter=" ")

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//img"),
                upper_boundary_selector=XPath("//div[@class='layout route cf']"),
                relative_urls=XPath("string((//head//link[@rel='stylesheet'])[1]/@href)"),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.publishers.ls.lesotho_times import LesothoTimesParser
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import Sitemap


class LS(metaclass=PublisherGroup):
    default_language = "en"

    LesothoTimes = Publisher(
        name="Lesotho Times",
        domain="https://lestimes.com/",
        parser=LesothoTimesParser,
        sources=[
            Sitemap(
                "https://lestimes.com/sitemap_index.xml",
                sitemap_filter=inverse(regex_filter("post-sitemap")),
                reverse=True,
            ),
        ],
    )



import datetime

from dateutil.rrule import MONTHLY, rrule

from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.publishers.es.abc import ABCParser
from fundus.publishers.es.el_diario import ElDiarioParser
from fundus.publishers.es.el_mundo import ElMundoParser
from fundus.publishers.es.el_pais import ElPaisParser
from fundus.publishers.es.la_vanguardia import LaVanguardiaParser
from fundus.publishers.es.mallorca_magazin import MallorcaMagazinParser
from fundus.publishers.es.mallorca_zeitung import MallorcaZeitungParser
from fundus.publishers.es.publico import PublicoParser
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap


class ES(metaclass=PublisherGroup):
    default_language = "es"

    ElPais = Publisher(
        name="El País",
        domain="https://elpais.com/",
        parser=ElPaisParser,
        sources=[RSSFeed("https://feeds.elpais.com/mrss-s/pages/ep/site/elpais.com/portada")]
        + [
            Sitemap(f"https://elpais.com/sitemaps/{d.year}/{str(d.month).zfill(2)}/sitemap_0.xml")
            for d in reversed(
                list(rrule(MONTHLY, dtstart=datetime.datetime(1976, 5, 1), until=datetime.datetime.now()))
            )
        ],
    )
    ElMundo = Publisher(
        name="El Mundo",
        domain="https://www.elmundo.es/",
        parser=ElMundoParser,
        sources=[
            RSSFeed("https://e00-elmundo.uecdn.es/elmundo/rss/portada.xml"),
            RSSFeed("https://e00-elmundo.uecdn.es/elmundo/rss/espana.xml"),
        ],
    )
    MallorcaMagazin = Publisher(
        name="Mallorca Magazin",
        domain="https://www.mallorcamagazin.com/",
        parser=MallorcaMagazinParser,
        sources=[
            NewsMap("https://www.mallorcamagazin.com/googlenews.xml", languages={"de"}),
            Sitemap(
                "https://www.mallorcamagazin.com/nachrichten/sitemapIndex.xml",
                reverse=True,
                languages={"de"},
            ),
        ],
    )
    ABC = Publisher(
        name="ABC",
        domain="https://www.abc.es/",
        parser=ABCParser,
        sources=[
            NewsMap("https://www.abc.es/sitemap.xml"),
            RSSFeed("https://www.abc.es/rss/2.0/espana/"),
            RSSFeed("https://www.abc.es/rss/2.0/portada/"),
        ],
    )
    LaVanguardia = Publisher(
        name=" La Vanguardia",
        domain="https://www.lavanguardia.com/",
        parser=LaVanguardiaParser,
        sources=[
            NewsMap("https://www.lavanguardia.com/sitemap-google-news.xml"),
            NewsMap("https://www.lavanguardia.com/sitemap-news-agencias.xml"),
            RSSFeed("https://www.lavanguardia.com/rss/home.xml"),
            RSSFeed("https://www.lavanguardia.com/rss/internacional.xml"),
        ]
        + [
            Sitemap(f"https://www.lavanguardia.com/sitemap-noticias-{d.year}{str(d.month).zfill(2)}.xml.gz")
            for d in reversed(
                list(rrule(MONTHLY, dtstart=datetime.datetime(2019, 1, 1), until=datetime.datetime.now()))
            )
        ],
    )

    MallorcaZeitung = Publisher(
        name="Mallorca Zeitung",
        domain="https://www.mallorcazeitung.es/",
        parser=MallorcaZeitungParser,
        sources=[
            NewsMap("https://www.mallorcazeitung.es/sitemap_google_news_8d82b.xml", languages={"de"}),
            RSSFeed("https://www.mallorcazeitung.es/rss/section/28000", languages={"de"}),
        ],
    )
    ElDiario = Publisher(
        name="elDiario.es",
        domain="https://www.eldiario.es/",
        parser=ElDiarioParser,
        sources=[
            RSSFeed("https://www.eldiario.es/rss/"),
            NewsMap("https://www.eldiario.es/sitemap_google_news_25b87.xml"),
            Sitemap(
                "https://www.eldiario.es/sitemap_index_25b87.xml",
                sitemap_filter=inverse(regex_filter("sitemap_contents")),
            ),
        ],
    )
    Publico = Publisher(
        name="Publico",
        domain="https://www.publico.es/",
        parser=PublicoParser,
        sources=[
            Sitemap(f"https://www.publico.es/sitemap-noticias-{d.year}{str(d.month).zfill(2)}.xml")
            for d in reversed(
                list(rrule(MONTHLY, dtstart=datetime.datetime(2020, 1, 1), until=datetime.datetime.now()))
            )
        ],
    )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_nodes_to_text,
    image_extraction,
    normalize_whitespace,
)


class LesEchosParser(ParserProxy):
    class V1(BaseParser):
        _summary_selector = CSSSelector("article header > p")
        _subheadline_selector = CSSSelector("article div.post-paywall > h3")

        _bloat_regex_ = r"^\s*Pour ne rien rater de l'actualité politique"

        _paragraph_selector = XPath(
            f'//article //div[contains(@class, "post-paywall")] /p[not(re:test(string(), "{_bloat_regex_}"))]',
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )

        _topic_selector = CSSSelector("header div.sc-108qdzy-3 div.sc-108qdzy-2 > div")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            if title := self.precomputed.meta.get("og:title"):
                return normalize_whitespace(title)
            return None

        @attribute
        def topics(self) -> List[str]:
            topic_nodes = self._topic_selector(self.precomputed.doc)
            return [normalize_whitespace(text) for text in generic_nodes_to_text(topic_nodes)]

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(
                self.precomputed.meta.get("article:published_time") or self.precomputed.ld.bf_search("datePublished")
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                author_selector=re.compile(r"\((?P<credits>.*?)\)$"),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

from .anadolu_ajansi import AnadoluAjansiParser
from .haberturk import HaberturkParser
from .ntvtr import NTVTRParser


class TR(metaclass=PublisherGroup):
    default_language = "tr"
    Haberturk = Publisher(
        name="Haberturk",
        domain="https://www.haberturk.com/",
        parser=HaberturkParser,
        sources=[
            Sitemap(
                "https://www.haberturk.com/sitemap.xml",
                sitemap_filter=inverse(regex_filter("news|special|posts|ozel_icerikler")),
                reverse=True,
            ),
            NewsMap("https://www.haberturk.com/sitemap_google_news.xml"),
        ],
    )

    NTVTR = Publisher(
        name="NTVTR",
        domain="https://www.ntv.com.tr/",
        parser=NTVTRParser,
        sources=[
            RSSFeed("https://www.ntv.com.tr/gundem.rss"),
            NewsMap("https://www.ntv.com.tr/sitemaps/news-sitemap.xml"),
            Sitemap("https://www.ntv.com.tr/sitemaps", sitemap_filter=regex_filter("news-sitemap.xml")),
        ],
    )

    AnadoluAjansi = Publisher(
        name="Anadolu Ajansı",
        domain="https://www.aa.com.tr/",
        parser=AnadoluAjansiParser,
        sources=[
            RSSFeed("https://www.aa.com.tr/tr/rss/default?cat=guncel"),
            RSSFeed("https://www.aa.com.tr/tr/teyithatti/rss/news?cat=gazze"),
            RSSFeed("https://www.aa.com.tr/tr/teyithatti/rss/news?cat=politika"),
            RSSFeed("https://www.aa.com.tr/tr/teyithatti/rss/news?cat=aktuel"),
            RSSFeed("https://www.aa.com.tr/tr/teyithatti/rss/news?cat=kultur-sanat"),
            RSSFeed("https://www.aa.com.tr/tr/teyithatti/rss/news?cat=bilim-teknoloji"),
            RSSFeed("https://www.aa.com.tr/tr/teyithatti/rss/news?cat=blog"),
            RSSFeed("https://www.aa.com.tr/tr/teyithatti/rss/news?cat=teyit-sozlugu"),
            RSSFeed("https://www.aa.com.tr/tr/teyithatti/rss/news?cat=ekonomi"),
            RSSFeed("https://www.aa.com.tr/tr/teyithatti/rss/news?cat=0"),
            RSSFeed("https://www.aa.com.tr/tr/teyithatti/rss/video"),
        ],
    )



import re
from typing import Set

from fundus.publishers.at import AT
from fundus.publishers.au import AU
from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.publishers.be import BE
from fundus.publishers.ca import CA
from fundus.publishers.ch import CH
from fundus.publishers.cn import CN
from fundus.publishers.cz import CZ
from fundus.publishers.de import DE
from fundus.publishers.dk import DK
from fundus.publishers.es import ES
from fundus.publishers.fr import FR
from fundus.publishers.gl import GL
from fundus.publishers.il import IL
from fundus.publishers.ind import IND
from fundus.publishers.isl import ISL
from fundus.publishers.it import IT
from fundus.publishers.jp import JP
from fundus.publishers.kr import KR
from fundus.publishers.li import LI
from fundus.publishers.ls import LS
from fundus.publishers.lt import LT
from fundus.publishers.lu import LU
from fundus.publishers.mx import MX
from fundus.publishers.my import MY
from fundus.publishers.na import NA
from fundus.publishers.no import NO
from fundus.publishers.pl import PL
from fundus.publishers.pt import PT
from fundus.publishers.py import PY
from fundus.publishers.ru import RU
from fundus.publishers.tr import TR
from fundus.publishers.tw import TW
from fundus.publishers.tz import TZ
from fundus.publishers.uk import UK
from fundus.publishers.us import US
from fundus.publishers.za import ZA

__all__ = ["Publisher", "PublisherGroup"]


class PublisherCollectionMeta(PublisherGroup):
    def __new__(cls, name, bases, attributes):
        publishers: Set[str] = set()
        for attribute, value in attributes.items():
            # skipp dunder attributes
            if re.search(r"^__.*__$", attribute):
                continue
            elif isinstance(value, PublisherGroup):
                new_publishers: Set[str] = {publisher.__name__ for publisher in value}
                if duplicate_publishers := publishers & new_publishers:
                    raise ValueError(
                        f"Publisher(s) {', '.join(duplicate_publishers)!r} " f"already exists in collection {name!r}."
                    )
                publishers.update(new_publishers)
            elif isinstance(value, Publisher):
                if attribute in publishers:
                    raise ValueError(f"Publishers {value.name!r} already exists in collection {name!r}.")
                publishers.add(attribute)
            else:
                raise TypeError(
                    f"Got unexpected type {type(value)} with value {value!r} "
                    f"Only objects of type {Publisher.__name__!r} or {PublisherGroup.__name__!r} "
                    f"are allowed as class attributes"
                )

        return super().__new__(cls, name, bases, attributes)


class PublisherCollection(metaclass=PublisherCollectionMeta):
    at = AT
    au = AU
    be = BE
    ca = CA
    ch = CH
    cn = CN
    cz = CZ
    de = DE
    dk = DK
    es = ES
    fr = FR
    gl = GL
    il = IL
    ind = IND
    isl = ISL
    it = IT
    jp = JP
    kr = KR
    li = LI
    ls = LS
    lt = LT
    lu = LU
    mx = MX
    my = MY
    na = NA
    no = NO
    pl = PL
    pt = PT
    py = PY
    ru = RU
    tr = TR
    tw = TW
    tz = TZ
    uk = UK
    us = US
    za = ZA



from typing import Iterator


def iterate_all_subclasses(cls: type) -> Iterator[type]:
    subclasses = cls.__subclasses__()
    yield from subclasses
    for subclass in subclasses:
        yield from iterate_all_subclasses(subclass)



import inspect
import json
from dataclasses import asdict, fields, is_dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Sequence,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from typing_extensions import TypeAlias

_T = TypeVar("_T")
_M = TypeVar("_M", bound="DataclassSerializationMixin")
JSONVal: TypeAlias = Union[None, bool, str, float, int, Sequence["JSONVal"], Dict[str, "JSONVal"]]


def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False


def replace_keys_in_nested_dict(data: Dict[str, _T], transformation: Callable[[str], str]) -> Dict[str, _T]:
    """Recursively replace all keys in a nested dictionary with <transformation>.

    Args:
        data: The dictionary to transform
        transformation: The transformation to use

    Returns:
        The transformed dictionary
    """

    def process(element) -> Any:
        if isinstance(element, dict):
            # Apply transformation to keys and recurse into values
            return {transformation(k): process(v) for k, v in element.items()}
        elif isinstance(element, list):
            # Recursively apply to elements in a list
            return [process(i) for i in element]
        else:
            # Base case: return the value as is if it's not a dict or list
            return element

    return {transformation(k): process(v) for k, v in data.items()}


class DataclassSerializationMixin:
    def serialize(self) -> Dict[str, JSONVal]:
        if not is_dataclass(self):
            raise TypeError(f"{type(self).__name__!r} is not a dataclass")
        return asdict(self)  # type: ignore[arg-type]

    @classmethod
    def deserialize(cls: Type[_M], serialized: Dict[str, JSONVal]) -> _M:
        if not is_dataclass(cls):
            raise TypeError(f"{type(cls).__name__!r} is not a dataclass")

        # we use get_type_hints here to resolve forward references since we need the actual types
        # not the forwarded string reference
        annotations = get_type_hints(cls)

        for field in fields(cls):
            serialized[field.name] = _inner_deserialize(serialized[field.name], annotations[field.name])

        return cls(**serialized)  # type: ignore[return-value]


def _inner_deserialize(data, cls):
    if data is None:
        return None

    if inspect.isclass(cls) and issubclass(cls, DataclassSerializationMixin):
        return cls.deserialize(data)
    elif (origin := get_origin(cls)) is list:
        item_type = get_args(cls)[0]  # Assuming homogeneous lists
        return [_inner_deserialize(item, item_type) for item in data]
    elif origin is dict:
        key_type, value_type = cls.__args__
        return {_inner_deserialize(k, key_type): _inner_deserialize(v, value_type) for k, v in data.items()}
    elif origin is Union:
        for union_type in cls.__args__:
            if union_type is not None:
                try:
                    return _inner_deserialize(data, union_type)
                except TypeError:
                    continue
        return None
    else:
        return data



import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
)


class CNBCParser(ParserProxy):
    class V1(BaseParser):
        _subheadline_selector: CSSSelector = CSSSelector("div[data-module = 'ArticleBody'] > h2")
        _paragraph_selector: XPath = XPath("//div[@data-module='ArticleBody'] / div[@class='group'] / p[text()]")
        _key_points_selector: CSSSelector = CSSSelector("div.RenderKeyPoints-list li")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            body: ArticleBody = extract_article_body_with_selector(
                self.precomputed.doc,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )
            return body

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.xpath_search("NewsArticle/author"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.xpath_search("NewsArticle/datePublished", scalar=True))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.xpath_search("NewsArticle/headline", scalar=True)

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute(validate=False)
        def key_points(self) -> List[str]:
            return [key_point.text_content() for key_point in self._key_points_selector(self.precomputed.doc)]

        """
        CNBC uses unconventional image loading, which is not supported at the time
        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//h1[@class='ArticleHeader-headline']"),
                image_selector=XPath("//div[@class='InlineImage-wrapper']//img"),
                caption_selector=XPath("./ancestor::div[@class='InlineImage-wrapper']//div[@class='InlineImage-imageEmbedCaption']"),
                author_selector=XPath(
                    "./ancestor::div[@class='InlineImage-wrapper']//div[@class='InlineImage-imageEmbedCredit']")
            )
        """



from datetime import datetime
from typing import List, Optional

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    image_extraction,
)


class TheGatewayPunditParser(ParserProxy):
    class V1(BaseParser):
        _related = r"^Click\s$"
        _paragraph_selector = XPath(
            f"(//div[@class='entry-content'] | //div[@class='entry-content']/blockquote[not(@class='twitter-tweet')]) "
            f"/p[not(child::img or child::script or re:test(text(), '{_related}')) and text()]",
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.xpath_search("Article/author"))

        @attribute
        def publishing_date(self) -> Optional[datetime]:
            return generic_date_parsing(self.precomputed.meta.get("article:published_time"))

        @attribute
        def title(self) -> Optional[str]:
            if (title := self.precomputed.meta.get("og:title")) is not None:
                title = title.split("|")[0].strip()
            return title

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//div[@class='entry-content']//img"),
                author_selector=XPath("./ancestor::figure//figcaption"),
            )



import datetime
import re
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    image_extraction,
)


class TagesschauParser(ParserProxy):
    class V1(BaseParser):
        _paragraph_selector = XPath("//article/p[position() > 1]")
        _summary_selector = XPath("//article/p[1]")
        _subheadline_selector = XPath("//article/h2")
        _author_selector = XPath('string(//div[contains(@class, "authorline__author")])')
        _topic_selector = CSSSelector("div.meldungsfooter .taglist a")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            if raw_author_string := self._author_selector(self.precomputed.doc):
                cleaned_author_string = re.sub(r"^Von |, ARD[^\s,]*", "", raw_author_string)
                return generic_author_parsing(cleaned_author_string)
            else:
                return []

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("og:title")

        @attribute
        def topics(self) -> List[str]:
            topic_nodes = self._topic_selector(self.precomputed.doc)
            return [node.text_content() for node in topic_nodes]

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath(
                    "//*[not(self::div and @class='teaser-absatz__image')]/div[@class='ts-picture__wrapper']//img"
                ),
                alt_selector=XPath("./@title"),
                author_selector=re.compile(r"\|(?P<credits>.+)"),
                caption_selector=XPath("./ancestor::div[contains(@class, 'absatzbild ')]"),
                lower_boundary_selector=self._topic_selector,
            )



import pytest

from fundus import NewsMap, RSSFeed, Sitemap


class TestCollection:
    def test_len(
        self,
        empty_publisher_group,
        group_with_empty_publisher_subgroup,
        group_with_valid_publisher_subgroup,
        group_with_two_valid_publisher_subgroups,
    ):
        assert len(empty_publisher_group) == 0
        assert len(group_with_empty_publisher_subgroup) == 0
        assert len(group_with_valid_publisher_subgroup) == 1
        assert len(group_with_two_valid_publisher_subgroups) == 2

    def test_iter_empty_group(self, empty_publisher_group):
        assert list(empty_publisher_group) == []

    def test_iter_group_with_empty_publisher_subgroup(self, group_with_empty_publisher_subgroup):
        assert list(group_with_empty_publisher_subgroup) == []

    def test_iter_group_with_publisher_subgroup(self, group_with_valid_publisher_subgroup):
        assert list(group_with_valid_publisher_subgroup) == [group_with_valid_publisher_subgroup.pub.value]

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_supports(self, publisher_group_with_news_map, publisher_group_with_languages):
        assert publisher_group_with_news_map.value.supports(source_types=[NewsMap])
        assert not publisher_group_with_news_map.value.supports(source_types=[Sitemap])
        assert not publisher_group_with_news_map.value.supports(source_types=[RSSFeed])
        assert publisher_group_with_languages.eng.supports(languages=["en"])
        assert not publisher_group_with_languages.eng.supports(languages=["es"])

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_search(
        self, publisher_group_with_news_map, proxy_with_two_versions_and_different_attrs, publisher_group_with_languages
    ):
        parser_proxy = proxy_with_two_versions_and_different_attrs()

        # monkey pathing publisher enums parser
        publisher_group_with_news_map.value.parser = parser_proxy

        later, earlier = parser_proxy.attribute_mapping.values()

        assert len(publisher_group_with_news_map.search(later.names, source_types=[NewsMap])) == 1
        assert len(publisher_group_with_news_map.search(later.names, source_types=[RSSFeed, Sitemap])) == 0
        assert len(publisher_group_with_news_map.search(later.names, source_types=[NewsMap, RSSFeed])) == 1

        # check that only latest version is supported with search
        assert len(publisher_group_with_news_map.search(later.names)) == 1
        assert len(publisher_group_with_news_map.search(earlier.names)) == 0

        assert len(publisher_group_with_languages.search(languages=["en", "en"])) == 1
        assert len(publisher_group_with_languages.search(languages=["de"])) == 1
        assert len(publisher_group_with_languages.search(languages=["en", "de"])) == 2

        assert len(publisher_group_with_languages.search(languages=["en"], source_types=[NewsMap])) == 1
        assert len(publisher_group_with_languages.search(languages=["en"], source_types=[Sitemap])) == 0

        assert len(publishers := publisher_group_with_languages.search(languages=["es"])) == 1
        assert len(publishers[0].source_mapping) == 2

        assert len(publishers := publisher_group_with_languages.search(languages=["ind"])) == 1
        assert len(publishers[0].source_mapping) == 1

        assert len(publishers := publisher_group_with_languages.search(languages=["pl"])) == 1
        assert len(publishers[0].source_mapping) == 1

        assert len(publishers := publisher_group_with_languages.search(languages=["es", "ind"])) == 1
        assert len(publishers[0].source_mapping) == 3

        assert len(publishers := publisher_group_with_languages.search(languages=["pl", "ind"])) == 1
        assert len(publishers[0].source_mapping) == 2

        assert len(publishers := publisher_group_with_languages.search(languages=["es"], source_types=[RSSFeed])) == 1
        assert len(publishers[0].source_mapping) == 1

        assert len(publishers := publisher_group_with_languages.search(languages=["es"], source_types=[NewsMap])) == 1
        assert len(publishers[0].source_mapping) == 1

        assert len(publisher_group_with_languages.search(languages=["es"], source_types=[Sitemap])) == 0

        assert (
            len(publishers := publisher_group_with_languages.search(languages=["es", "ind"], source_types=[Sitemap]))
            == 1
        )
        assert len(publishers[0].source_mapping) == 1

        with pytest.raises(ValueError):
            publisher_group_with_news_map.search()

        with pytest.raises(ValueError):
            publisher_group_with_news_map.search([])

        with pytest.raises(ValueError):
            publisher_group_with_news_map.search([], [])

    def test_publisher_group_with_subgroup_string_representation(self, group_with_two_valid_publisher_subgroups):
        representation = str(group_with_two_valid_publisher_subgroups)
        assertion = (
            "<GroupWithTwoValidatePublisherSubGroups: 2>"
            "\n\t<PubGroupNews: 1>"
            "\n\t\ttest_pub"
            "\n\t<PubGroupSitemap: 1>"
            "\n\t\ttest_pub"
        )
        assert representation == assertion

    def test_publisher_group_with_publisher_string_representation(self, publisher_group_with_news_map):
        representation = str(publisher_group_with_news_map)
        assertion = "<PubGroup: 1>\n\ttest_pub"
        assert representation == assertion



import threading
from collections import defaultdict
from typing import Any, Dict, Optional, Union

from fundus.logging import create_logger

logger = create_logger(__name__)


class EventDict:
    """A thread-safe event dictionary.

    Events are registered by name and stored per thread in a dictionary, using the
    thread's identifier as the key. For example, calling `register_event("stop")`
    registers a "stop" event for the current thread's identifier.

    To enhance usability, threads can be assigned aliases. Calling
    `register_event("stop", "BR")` registers the "stop" event (if it is not already
    registered) for the current thread and automatically creates an alias mapping
    "BR" to the thread's identifier.
    """

    def __init__(self):
        self._events: Dict[int, Dict[str, threading.Event]] = defaultdict(dict)
        self._aliases: Dict[Any, int] = {}
        self._lock = threading.Lock()

    @staticmethod
    def _get_identifier() -> int:
        return threading.get_ident()

    def _resolve(self, key: Union[int, str, None]) -> int:
        """Resolves a given key to a thread identifier

        Should only be used within a Lock!

        Args:
            key: Key to resolve

        Returns:
            Resolved thread identifier
        """
        if key is None:
            return self._get_identifier()
        if isinstance(key, int):
            return key
        return self._aliases[key]

    def _alias(self, alias: str, key: Optional[int] = None):
        self._aliases[alias] = key if key else self._get_identifier()
        logger.debug(f"Registered alias {alias} -> {self._aliases[alias]}")

    def register_event(self, event: str, key: Union[int, str, None] = None):
        with self._lock:
            if isinstance(key, str) and key not in self._aliases:
                self._alias(key)
            if (resolved := self._resolve(key)) not in self._events:
                self._events[resolved][event] = threading.Event()
                logger.debug(f"Registered event {event!r} for {resolved}")

    def set_event(self, event: str, key: Union[int, str, None] = None):
        with self._lock:
            self._events[self._resolve(key)][event].set()
            logger.debug(f"Set event {event!r} for {self._resolve(key)}")

    def clear_event(self, event: str, key: Union[int, str, None] = None):
        with self._lock:
            self._events[self._resolve(key)][event].clear()
            logger.debug(f"Cleared event {event!r} for {self._resolve(key)}")

    def set_for_all(self, event: Optional[str] = None):
        """Set <event> for all registered keys

        If <event> is None, all events for every registered key will be set.
        Args:
            event: The event to set. Defaults to None.

        Returns:
            None
        """
        with self._lock:
            for events in self._events.values():
                if event is not None and event in events:
                    events[event].set()
                else:
                    for flag in events.values():
                        flag.set()

    def clear_for_all(self, event: Optional[str] = None):
        """Clear <event> for all registered keys

        If <event> is None, all events for every registered key will be cleared.
        Args:
            event: The event to clear. Defaults to None.

        Returns:
            None
        """
        with self._lock:
            for events in self._events.values():
                if event is not None and event in events:
                    events[event].clear()
                else:
                    for flag in events.values():
                        flag.clear()

    def is_event_set(self, event: str, key: Union[int, str, None] = None) -> bool:
        with self._lock:
            return self._events[self._resolve(key)][event].is_set()

    def alias(self, alias: str, key: Optional[int] = None):
        with self._lock:
            self._alias(alias, key)

    def remove_alias(self, alias: str):
        with self._lock:
            self._aliases.pop(alias, None)

    def get(self, event: str, key: Optional[Union[int, str, None]] = None) -> threading.Event:
        with self._lock:
            return self._events[self._resolve(key)][event]


__EVENTS__: EventDict = EventDict()



import pathlib

from langdetect import DetectorFactory

from fundus.publishers import PublisherCollection
from fundus.scraping.article import Article
from fundus.scraping.crawler import CCNewsCrawler, Crawler, CrawlerBase
from fundus.scraping.filter import Requires
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

__module_path__ = pathlib.Path(__file__).parent
__development_base_path__ = __module_path__.parents[1]

__all__ = [
    "CrawlerBase",
    "Crawler",
    "CCNewsCrawler",
    "PublisherCollection",
    "Requires",
    "RSSFeed",
    "Sitemap",
    "NewsMap",
    "Article",
]

# make language detection deterministic https://pypi.org/project/langdetect/
DetectorFactory.seed = 0



import datetime
import re
from typing import List, Optional, Pattern

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
    normalize_whitespace,
)


class APNewsParser(ParserProxy):
    class V1(BaseParser):
        VALID_UNTIL = datetime.date(2023, 7, 10)
        _author_selector: XPath = XPath(f"{CSSSelector('div.CardHeadline').path}/span/span[1]")
        _subheadline_selector = XPath("//div[@data-key = 'article']/h2[not(text()='___')]")
        _paragraph_selector = XPath("//div[@data-key = 'article']/p")

        _topic_bloat_pattern: Pattern[str] = re.compile(r"state wire| news|^.{1}$", flags=re.IGNORECASE)

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
            )

        @attribute
        def authors(self) -> List[str]:
            # AP News does not have all the article's authors listed in the linked data.
            # Therefore, we try to parse the article's authors from the document.
            try:
                # Example: "By AUTHOR1, AUTHOR2 and AUTHOR3"
                author_string: str = normalize_whitespace(self._author_selector(self.precomputed.doc)[0].text_content())
                author_string = re.sub(r"^By ", "", author_string)
            except IndexError:
                # Fallback to the generic author parsing from the linked data.
                return generic_author_parsing(self.precomputed.ld.xpath_search("NewsArticle/author"))

            return generic_author_parsing(author_string)

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.xpath_search("NewsArticle/datePublished", scalar=True))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.xpath_search("NewsArticle/headline", scalar=True)

        @attribute
        def topics(self) -> List[str]:
            return [
                topic
                for topic in generic_topic_parsing(self.precomputed.meta.get("keywords"))
                if not re.search(self._topic_bloat_pattern, topic)
            ]

        # unfortunately we would need to render the site first before parsing images for this version

    class V1_1(V1):
        VALID_UNTIL = datetime.date.today()

        _author_selector = CSSSelector("div.Page-authors")
        _subheadline_selector = XPath("//div[contains(@class, 'RichTextStoryBody')] /h2[not(text()='___')]")
        _paragraph_selector = XPath(
            "//div[contains(@class, 'RichTextStoryBody')] "
            "/p[not(preceding-sibling::*[1][self::h2 and text()='___'])]"
            # only p-elements not directly following h2 elements with text() = '___'
        )

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=XPath("//*[self::figure or @class='CarouselSlide']//img"),
                caption_selector=XPath(
                    "./ancestor::figure//figcaption | "
                    "./ancestor::div[@class='CarouselSlide']//span[@class='CarouselSlide-infoDescription']"
                ),
                upper_boundary_selector=XPath("//div[@class='Page-content' or @class='Body']"),
                lower_boundary_selector=CSSSelector("footer.Page-footer"),
                author_selector=re.compile(r"\s*\((?P<credits>.*)\)$"),
            )



import datetime
import re
from typing import List, Optional, Pattern

from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, ParserProxy, attribute
from fundus.parser.data import Image
from fundus.parser.utility import (
    apply_substitution_pattern_over_list,
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class FunkeParser(ParserProxy):
    class V1(BaseParser):
        VALID_UNTIL = datetime.date(2025, 8, 24)

        _author_substitution_pattern: Pattern[str] = re.compile(r"FUNKE Mediengruppe|.*dpa(:|-infocom).*|^red$")
        _paragraph_selector = XPath(
            "//div[@class='article-body']//p[not(not(text()) or @rel='author' or em[@class='print'] or @class)]"
        )
        _summary_selector = XPath("//div[@class='article-body']/p[contains(@class, 'font-sans')]")
        _subheadline_selector = XPath(
            "//div[@class='article-body']//h3[not("
            "contains(text(), 'Alle Artikel der Serie')"
            " or contains(text(), 'Mehr zum Thema')"
            " or contains(text(), 'weitere Videos')"
            " or contains(text(), 'Auch interessant')"
            " or contains(text(), 'Weitere News')"
            " or @class)]"
        )
        _topics_selector = XPath("//div[@class='not-prose  mb-4 mx-5 font-sans']/ul/li")
        _image_selector = XPath("//img[not(contains(@class, 'rounded-full'))]")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.ld.bf_search("headline")

        @attribute
        def topics(self) -> List[str]:
            if topics := generic_topic_parsing(self.precomputed.meta.get("news_keywords")):
                return topics
            else:
                pass
            return [
                re.sub(r"\s*–.+", "", node.text_content()).strip()
                for node in self._topics_selector(self.precomputed.doc)
            ]

        @attribute
        def authors(self) -> List[str]:
            return apply_substitution_pattern_over_list(
                generic_author_parsing(self.precomputed.ld.bf_search("author")), self._author_substitution_pattern
            )

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=self._image_selector,
                author_selector=re.compile(r"©(?P<credits>.*)"),
            )

    class V1_1(V1):
        VALID_UNTIL = datetime.date.today()

        _paragraph_selector = XPath(
            "//div[contains(@class,'article-body')]"
            "/p[contains(@class,'expressive-copy-lg-body') and not(contains(text(), '>>')) and string-length(text()) > 10]|"
            "//div[contains(@class,'article-body')]//ul/li[string-length(text())>10 and not(a or article)]"
        )
        _subheadline_selector = XPath(
            "//div[contains(@class,'article-body')]//h3[not("
            "contains(text(), 'Alle Artikel der Serie')"
            " or contains(text(), 'Mehr zum Thema')"
            " or contains(text(), 'weitere Videos')"
            " or contains(text(), 'Auch interessant')"
            " or contains(text(), 'Weitere News')"
            " or not(contains(@class, 'expressive-heading-xl'))"
            " or following-sibling::*[1][self::ul])]"
        )
        _summary_selector = XPath("//div[contains(@class, 'expressive-copy-lg')]")
        _image_selector = XPath(
            "//img[contains(@class, 'lg:aspect-[16/9]') or not(contains(@class, 'aspect-[1/1]'))] | //figure/picture"
        )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

from .kommersant import KommersantParser


class RU(metaclass=PublisherGroup):
    default_language = "ru"

    Kommersant = Publisher(
        name="Kommersant",
        domain="https://www.kommersant.ru/",
        parser=KommersantParser,
        sources=[
            RSSFeed("https://www.kommersant.ru/RSS/news.xml"),
            Sitemap(
                "https://www.kommersant.ru/sitemaps/all_docs_sitemap.xml",
                reverse=True,
            ),
            NewsMap("https://www.kommersant.ru/sitemaps/sitemap_news.xml"),
        ],
        suppress_robots=True,
    )



import re
from datetime import datetime
from typing import List, Optional

from lxml.cssselect import CSSSelector
from lxml.etree import XPath, tostring
from lxml.html import document_fromstring

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
    transform_breaks_to_paragraphs,
)


class IlGiornaleParser(ParserProxy):
    class V1(BaseParser):
        # Selectors for article body parts
        _paragraph_selector = XPath(
            "//div[contains(@class, 'typography--content')]//p[text() or strong or em] | //div[@class='banner banner--spaced-block banner-evo' and (text() or em or strong)]"
        )
        _subheadline_selector = CSSSelector("div.typography--content h2:not([class])")
        _summary_selector = CSSSelector("p.article__abstract, div.article__abstract")
        _image_selector = XPath(
            "//div[contains(@class, 'article__media')]//img | //section[contains(@class, 'article__content')]//img"
        )

        @attribute
        def title(self) -> Optional[str]:
            # First try JSON-LD
            title = self.precomputed.ld.xpath_search("//NewsArticle/headline", scalar=True)
            if title:
                return str(title)
            # Fallback to meta tags
            return self.precomputed.meta.get("og:title")

        @attribute
        def authors(self) -> List[str]:
            # Extract authors from schema.org NewsArticle data
            authors = self.precomputed.ld.xpath_search("//NewsArticle/author")
            if authors:
                return generic_author_parsing(authors)
            return []

        @attribute
        def publishing_date(self) -> Optional[datetime]:
            # Try JSON-LD first
            date_str = self.precomputed.ld.xpath_search("//NewsArticle/datePublished", scalar=True)
            if not date_str:
                # Fallback to meta tags
                date_str = self.precomputed.meta.get("article:published_time")
            return generic_date_parsing(date_str)

        @attribute
        def body(self) -> Optional[ArticleBody]:
            # Clean up HTML by removing ads and handling em/strong/cite tags
            html_string = tostring(self.precomputed.doc).decode("utf-8")
            html_string = re.sub(r"</?(em|strong|cite)>", "", html_string)
            html_string = re.sub(r"<!-- EVOLUTION ADV -->", "", html_string)
            doc = document_fromstring(html_string)

            # Transform br tags to paragraphs for better structure
            doc = transform_breaks_to_paragraphs(doc)

            # Extract article body using utility function
            return extract_article_body_with_selector(
                doc,
                paragraph_selector=self._paragraph_selector,
                subheadline_selector=self._subheadline_selector,
                summary_selector=self._summary_selector,
            )

        @attribute
        def topics(self) -> List[str]:
            # Try to get topics from keywords
            keywords = self.precomputed.ld.bf_search("keywords")
            if keywords:
                return generic_topic_parsing(keywords)

            # Fallback to articleSection
            section = self.precomputed.ld.xpath_search("//NewsArticle/articleSection", scalar=True)
            if section:
                return generic_topic_parsing([section])

            return []

        @attribute
        def images(self) -> List[Image]:
            # Extract images using the utility function
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                image_selector=self._image_selector,
                caption_selector=XPath(".//figcaption/text()"),
            )



import datetime
import re
from typing import List, Optional, Union

from lxml.cssselect import CSSSelector
from lxml.etree import XPath

from fundus.parser import ArticleBody, BaseParser, Image, ParserProxy, attribute
from fundus.parser.utility import (
    extract_article_body_with_selector,
    generic_author_parsing,
    generic_date_parsing,
    generic_topic_parsing,
    image_extraction,
)


class NDRParser(ParserProxy):
    class V1(BaseParser):
        VALID_UNTIL = datetime.date(2025, 6, 16)

        _paragraph_selector = XPath(
            "//div[@class='modulepadding copytext']/p[not(@class='textauthor' or @class='preface')] "
            "| //div[@class='modulepadding copytext']/ul/li"
        )
        _summary_selector: Union[XPath, CSSSelector] = CSSSelector(".preface")
        _subheadline_selector: Union[XPath, CSSSelector] = CSSSelector("article .modulepadding > h2")

        @attribute
        def body(self) -> Optional[ArticleBody]:
            return extract_article_body_with_selector(
                self.precomputed.doc,
                summary_selector=self._summary_selector,
                subheadline_selector=self._subheadline_selector,
                paragraph_selector=self._paragraph_selector,
            )

        @attribute
        def topics(self) -> List[str]:
            return generic_topic_parsing(self.precomputed.meta.get("keywords"))

        @attribute
        def publishing_date(self) -> Optional[datetime.datetime]:
            return generic_date_parsing(self.precomputed.ld.bf_search("datePublished"))

        @attribute
        def authors(self) -> List[str]:
            return generic_author_parsing(self.precomputed.ld.bf_search("author"))

        @attribute
        def title(self) -> Optional[str]:
            return self.precomputed.meta.get("title")

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//div[@id='page']"),
                image_selector=XPath(
                    "//div[@id='page']//*[(self::div and not(@class='teaserimage')) or (self::a and @class='zoomimage')]/div[contains(@class,'image-container')]//picture//img"
                ),
                relative_urls=XPath("string(//link[@rel='canonical']/@href)"),
                caption_selector=XPath("./ancestor::div[contains(@class,'contentimage')]//span[@class='caption']"),
                author_selector=re.compile(r"(?i)©\s*(ndr)?\s*(foto)?:?\s*(?P<credits>.+)"),
            )

    class V1_1(V1):
        VALID_UNTIL = datetime.date.today()

        _paragraph_selector = XPath("//article/p[not(@class='textauthor')] | //article/ul/li | //article/blockquote")
        _subheadline_selector = XPath("//article/h2")
        _summary_selector = XPath("//header/p[@class='preface']")

        _bloat_keywords = ["hh", "regionalmeldungen", "News", "kurzmeldungen"]

        @attribute
        def topics(self) -> List[str]:
            return [
                topic
                for topic in generic_topic_parsing(self.precomputed.meta.get("keywords"))
                if topic not in self._bloat_keywords
            ]

        @attribute
        def images(self) -> List[Image]:
            return image_extraction(
                doc=self.precomputed.doc,
                paragraph_selector=self._paragraph_selector,
                upper_boundary_selector=XPath("//article"),
                relative_urls=XPath("string(//link[@rel='canonical']/@href)"),
                image_selector=XPath(
                    "//article//div[contains(@class,'contentimage') or contains(@class, 'herocontainer')]//picture//img"
                ),
                caption_selector=XPath(
                    "./ancestor::div[contains(@class,'contentimage')]//span[contains(@class, 'caption')]"
                ),
                author_selector=re.compile(r"(?i)©\s*(ndr)?\s*(foto)?:?\s*(?P<credits>.+)"),
            )



from fundus.publishers.base_objects import Publisher, PublisherGroup
from fundus.scraping.filter import inverse, regex_filter
from fundus.scraping.url import NewsMap, RSSFeed, Sitemap

from .luxemburger_wort import LuxemburgerWortParser
from .tageblatt import TageblattParser


class LU(metaclass=PublisherGroup):
    default_language = "de"

    Tageblatt = Publisher(
        name="Tageblatt",
        domain="https://www.tageblatt.lu/",
        parser=TageblattParser,
        sources=[
            RSSFeed("https://www.tageblatt.lu/category/politik/feed/atom/"),
            RSSFeed("https://www.tageblatt.lu/category/meinung/feed/atom/"),
            RSSFeed("https://www.tageblatt.lu/category/nachrichten/feed/atom/"),
            RSSFeed("https://www.tageblatt.lu/category/wirtschaft/feed/atom/"),
            RSSFeed("https://www.tageblatt.lu/category/sport/feed/atom/"),
            RSSFeed("https://www.tageblatt.lu/category/kultur/feed/atom/"),
            RSSFeed("https://www.tageblatt.lu/category/wissen/feed/atom/"),
            RSSFeed("https://www.tageblatt.lu/category/campus/feed/atom/"),
            RSSFeed("https://www.tageblatt.lu/category/magazin/feed/atom/"),
            RSSFeed("https://www.tageblatt.lu/category/auto/feed/atom/"),
            Sitemap(
                "https://www.tageblatt.lu/wp-sitemap.xml",
                sitemap_filter=inverse(regex_filter("posts-post")),
                reverse=True,
            ),
        ],
    )

    LuxemburgerWort = Publisher(
        name="Luxemburger Wort",
        domain="https://www.wort.lu/",
        parser=LuxemburgerWortParser,
        sources=[
            RSSFeed("https://www.wort.lu/de/rss"),
            Sitemap("https://www.wort.lu/sitemap.xml", reverse=True),
            NewsMap("https://www.wort.lu/sitemap-news.xml"),
        ],
    )
